"""
Tests for the buff system contrib
"""
from unittest.mock import Mock, call, patch

from evennia import DefaultObject, create_object
from evennia.contrib.rpg.buffs import buff
from evennia.utils import create
from evennia.utils.test_resources import EvenniaTest
from evennia.utils.utils import lazy_property

from .buff import BaseBuff, BuffableProperty, BuffHandler, Mod
from .samplebuffs import StatBuff


class _EmptyBuff(BaseBuff):
    key = "empty"
    pass


class _TestDivBuff(BaseBuff):
    key = "tdb"
    name = "tdb"
    flavor = "divverbuff"
    mods = [Mod("stat1", "div", 1)]


class _TestModBuff(BaseBuff):
    key = "tmb"
    name = "tmb"
    flavor = "modderbuff"
    maxstacks = 5
    mods = [Mod("stat1", "add", 10, 5), Mod("stat2", "mult", 0.5)]


class _TestModBuff2(BaseBuff):
    key = "tmb2"
    name = "tmb2"
    flavor = "modderbuff2"
    maxstacks = 1
    mods = [Mod("stat1", "mult", 1.0), Mod("stat1", "add", 10)]


class _TestTrigBuff(BaseBuff):
    key = "ttb"
    name = "ttb"
    flavor = "triggerbuff"
    triggers = ["test1", "test2"]

    def at_trigger(self, trigger: str, *args, **kwargs):
        if trigger == "test1":
            self.owner.db.triggertest1 = True
        if trigger == "test2":
            self.owner.db.triggertest2 = True


class _TestConBuff(BaseBuff):
    key = "tcb"
    name = "tcb"
    flavor = "condbuff"
    triggers = ["condtest"]

    def conditional(self, *args, **kwargs):
        return self.owner.db.cond1

    def at_trigger(self, trigger: str, attacker=None, defender=None, damage=0, *args, **kwargs):
        defender.db.att, defender.db.dmg = attacker, damage


class _TestComplexBuff(BaseBuff):
    key = "tcomb"
    name = "complex"
    flavor = "combuff"
    triggers = ["comtest", "complextest"]

    mods = [
        Mod("com1", "add", 0, 10),
        Mod("com1", "add", 15),
        Mod("com1", "mult", 2.0),
        Mod("com2", "add", 100),
    ]

    def conditional(self, cond=False, *args, **kwargs):
        return not cond

    def at_trigger(self, trigger: str, *args, **kwargs):
        if trigger == "comtest":
            self.owner.db.comtext = {"cond": True}
        else:
            self.owner.db.comtext = {}


class _TestTimeBuff(BaseBuff):
    key = "ttib"
    name = "ttib"
    flavor = "timerbuff"
    maxstacks = 1
    tickrate = 1
    duration = 5
    mods = [Mod("timetest", "add", 665)]

    def at_tick(self, initial=True, *args, **kwargs):
        self.owner.db.ticktest = True


class BuffableObject(DefaultObject):
    stat1 = BuffableProperty(10)

    @lazy_property
    def buffs(self) -> BuffHandler:
        return BuffHandler(self)

    def at_init(self):
        self.stat1, self.buffs
        return super().at_init()


class TestBuffsAndHandler(EvenniaTest):
    "This tests a number of things about buffs."

    def setUp(self):
        super().setUp()
        self.testobj = create.create_object(BuffableObject, key="testobj")

    def tearDown(self):
        """done after every test_* method below"""
        self.testobj.buffs.clear()
        del self.testobj
        super().tearDown()

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_addremove(self):
        """tests adding and removing buffs"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        # add
        handler.add(_TestModBuff, to_cache={"cachetest": True})
        handler.add(_TestTrigBuff)
        self.assertEqual(self.testobj.db.buffs["tmb"]["ref"], _TestModBuff)
        self.assertTrue(self.testobj.db.buffs["tmb"].get("cachetest"))
        self.assertFalse(self.testobj.db.buffs["ttb"].get("cachetest"))
        # has
        self.assertTrue(handler.has(_TestModBuff))
        self.assertTrue(handler.has("tmb"))
        self.assertFalse(handler.has(_EmptyBuff))
        # remove
        handler.remove("tmb")
        self.assertFalse(self.testobj.db.buffs.get("tmb"))
        # remove stacks
        handler.add(_TestModBuff, stacks=3)
        handler.remove("tmb", stacks=3)
        self.assertFalse(self.testobj.db.buffs.get("tmb"))
        # remove by type
        handler.add(_TestModBuff)
        handler.remove_by_type(_TestModBuff)
        self.assertFalse(self.testobj.db.buffs.get("tmb"))
        # remove by buff instance
        handler.add(_TestModBuff)
        handler.all["tmb"].remove()
        self.assertFalse(self.testobj.db.buffs.get("tmb"))
        # remove by source
        handler.add(_TestModBuff)
        handler.remove_by_source(None)
        self.assertFalse(self.testobj.db.buffs.get("tmb"))
        # remove by cachevalue
        handler.add(_TestModBuff)
        handler.remove_by_cachevalue("failure", True)
        self.assertTrue(self.testobj.db.buffs.get("tmb"))
        # remove all
        handler.add(_TestModBuff)
        handler.clear()
        self.assertFalse(self.testobj.buffs.all)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_getters(self):
        """tests all built-in getters"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        handler.add(_TestModBuff, source=self.obj2)
        handler.add(_TestTrigBuff, to_cache={"ttbcache": True})
        # normal getter
        self.assertTrue(isinstance(handler.get("tmb"), _TestModBuff))
        # stat getters
        self.assertTrue(isinstance(handler.get_by_stat("stat1")["tmb"], _TestModBuff))
        self.assertFalse(handler.get_by_stat("nullstat"))
        # trigger getters
        self.assertTrue("ttb" in handler.get_by_trigger("test1").keys())
        self.assertFalse("ttb" in handler.get_by_trigger("nulltrig").keys())
        # type getters
        self.assertTrue("tmb" in handler.get_by_type(_TestModBuff))
        self.assertFalse("tmb" in handler.get_by_type(_EmptyBuff))
        # source getter
        self.assertTrue("tmb" in handler.get_by_source(self.obj2))
        self.assertFalse("ttb" in handler.get_by_source(self.obj2))
        # cachevalue getter
        self.assertFalse("tmb" in handler.get_by_cachevalue("ttbcache"))
        self.assertTrue("ttb" in handler.get_by_cachevalue("ttbcache"))
        self.assertTrue("ttb" in handler.get_by_cachevalue("ttbcache", True))

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_details(self):
        """tests that buff details like name and flavor are correct; also test modifier viewing"""
        handler: BuffHandler = self.testobj.buffs
        handler.add(_TestModBuff)
        handler.add(_TestTrigBuff)
        self.assertEqual(handler.get("tmb").flavor, "modderbuff")
        self.assertEqual(handler.get("ttb").name, "ttb")
        mods = handler.view_modifiers("stat1")
        _testmods = {
            "add": {"total": 15, "strongest": 15},
            "mult": {"total": 0, "strongest": 0},
            "div": {"total": 0, "strongest": 0},
        }
        self.assertDictEqual(mods, _testmods)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_modify(self):
        """tests to ensure that values are modified correctly, and stack across mods"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        _stat1, _stat2 = 0, 10
        handler.add(_TestModBuff)
        # stat1 and 2 basic mods
        self.assertEqual(handler.check(_stat1, "stat1"), 15)
        self.assertEqual(handler.check(_stat2, "stat2"), 15)
        # checks can take any base value
        self.assertEqual(handler.check(_stat1, "stat2"), 0)
        self.assertEqual(handler.check(_stat2, "stat1"), 25)
        # change to base stat reflected in check
        _stat1 += 5
        self.assertEqual(handler.check(_stat1, "stat1"), 20)
        _stat2 += 10
        self.assertEqual(handler.check(_stat2, "stat2"), 30)
        # test stacking; single stack, multiple stack, max stacks
        handler.add(_TestModBuff)
        self.assertEqual(handler.check(_stat1, "stat1"), 25)
        handler.add(_TestModBuff, stacks=3)
        self.assertEqual(handler.check(_stat1, "stat1"), 40)
        handler.add(_TestModBuff, stacks=5)
        self.assertEqual(handler.check(_stat1, "stat1"), 40)
        # stat2 mod doesn't stack
        self.assertEqual(handler.check(_stat2, "stat2"), 30)
        # layers with second mod
        handler.add(_TestModBuff2)
        self.assertEqual(handler.check(_stat1, "stat1"), 100)
        self.assertEqual(handler.check(_stat2, "stat2"), 30)
        # apply only the strongest value
        self.assertEqual(handler.check(_stat1, "stat1", strongest=True), 80)
        # removing mod properly reduces value, doesn't affect other mods
        handler.remove_by_type(_TestModBuff)
        self.assertEqual(handler.check(_stat1, "stat1"), 30)
        self.assertEqual(handler.check(_stat2, "stat2"), 20)
        # divider mod test
        handler.add(_TestDivBuff)
        self.assertEqual(handler.check(_stat1, "stat1"), 15)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_trigger(self):
        """tests to ensure triggers correctly fire"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        handler.add(_TestTrigBuff)
        # trigger buffs
        handler.trigger("nulltest")
        self.assertFalse(self.testobj.db.triggertest1)
        self.assertFalse(self.testobj.db.triggertest2)
        handler.trigger("test1")
        self.assertTrue(self.testobj.db.triggertest1)
        self.assertFalse(self.testobj.db.triggertest2)
        handler.trigger("test2")
        self.assertTrue(self.testobj.db.triggertest1)
        self.assertTrue(self.testobj.db.triggertest2)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_context_conditional(self):
        """tests to ensure context is passed to buffs, and also tests conditionals"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        handler.add(_TestConBuff)
        self.testobj.db.cond1, self.testobj.db.att, self.testobj.db.dmg = False, None, 0
        # context to pass, containing basic event data and a little extra to be ignored
        _testcontext = {
            "attacker": self.obj2,
            "defender": self.testobj,
            "damage": 5,
            "overflow": 10,
        }
        # test negative conditional
        self.assertEqual(
            handler.get_by_type(_TestConBuff)["tcb"].conditional(**_testcontext), False
        )
        handler.trigger("condtest", _testcontext)
        self.assertEqual(self.testobj.db.att, None)
        self.assertEqual(self.testobj.db.dmg, 0)
        # test positive conditional + context passing
        self.testobj.db.cond1 = True
        self.assertEqual(handler.get_by_type(_TestConBuff)["tcb"].conditional(**_testcontext), True)
        handler.trigger("condtest", _testcontext)
        self.assertEqual(self.testobj.db.att, self.obj2)
        self.assertEqual(self.testobj.db.dmg, 5)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_complex(self):
        """tests a complex mod (conditionals, multiple triggers/mods)"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        self.testobj.db.comone, self.testobj.db.comtwo, self.testobj.db.comtext = 10, 0, {}
        handler.add(_TestComplexBuff)
        # stat checks work correctly and separately
        self.assertEqual(self.testobj.db.comtext, {})
        self.assertEqual(handler.check(self.testobj.db.comone, "com1"), 105)
        self.assertEqual(handler.check(self.testobj.db.comtwo, "com2"), 100)
        # stat checks don't happen if the conditional is true
        handler.trigger("comtest", self.testobj.db.comtext)
        self.assertEqual(self.testobj.db.comtext, {"cond": True})
        self.assertEqual(
            handler.get_by_type(_TestComplexBuff)["tcomb"].conditional(**self.testobj.db.comtext),
            False,
        )
        self.assertEqual(
            handler.check(self.testobj.db.comone, "com1", context=self.testobj.db.comtext), 10
        )
        self.assertEqual(
            handler.check(self.testobj.db.comtwo, "com2", context=self.testobj.db.comtext), 0
        )
        handler.trigger("complextest", self.testobj.db.comtext)
        self.assertEqual(
            handler.check(self.testobj.db.comone, "com1", context=self.testobj.db.comtext), 10
        )
        self.assertEqual(
            handler.check(self.testobj.db.comtwo, "com2", context=self.testobj.db.comtext), 0
        )
        # separate trigger follows different codepath
        self.testobj.db.comtext = {"cond": False}
        handler.trigger("complextest", self.testobj.db.comtext)
        self.assertEqual(self.testobj.db.comtext, {})
        self.assertEqual(
            handler.check(self.testobj.db.comone, "com1", context=self.testobj.db.comtext), 105
        )
        self.assertEqual(
            handler.check(self.testobj.db.comtwo, "com2", context=self.testobj.db.comtext), 100
        )

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay")
    def test_timing(self, mock_delay: Mock):
        """tests timing-related features, such as ticking and duration"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        mock_delay.side_effect = [None, handler.cleanup]
        handler.add(_TestTimeBuff)
        calls = [
            call(
                1,
                buff.tick_buff,
                handler=handler,
                buffkey="ttib",
                context={},
                initial=False,
                persistent=True,
            ),
            call(5, handler.cleanup, persistent=True),
        ]
        mock_delay.assert_has_calls(calls)
        self.testobj.db.timetest, self.testobj.db.ticktest = 1, False
        # test duration and ticking
        _instance = handler.get("ttib")
        self.assertTrue(_instance.ticking)
        self.assertEqual(_instance.duration, 5)
        _instance.at_tick()
        self.assertTrue(self.testobj.db.ticktest)
        # test duration modification and cleanup
        _instance.duration = 0
        self.assertEqual(handler.get("ttib").duration, 0)
        handler.cleanup()
        self.assertFalse(handler.get("ttib"), None)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_cacheattrlink(self):
        """tests the link between the instance attribute and the cache attribute"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        handler.add(_EmptyBuff)
        self.assertEqual(handler.buffcache["empty"]["duration"], -1)
        empty: _EmptyBuff = handler.get("empty")
        empty.duration = 30
        self.assertEqual(handler.buffcache["empty"]["duration"], 30)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_buffableproperty(self):
        """tests buffable properties"""
        # setup
        self.testobj.buffs.add(_TestModBuff)
        self.assertEqual(self.testobj.stat1, 25)
        self.testobj.buffs.remove("tmb")
        self.assertEqual(self.testobj.stat1, 10)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_stresstest(self):
        """tests large amounts of buffs, and related removal methods"""
        # setup
        for x in range(1, 20):
            self.testobj.buffs.add(_TestModBuff, key="test" + str(x))
            self.testobj.buffs.add(_TestTrigBuff, key="trig" + str(x))
        self.assertEqual(self.testobj.stat1, 295)
        self.testobj.buffs.trigger("test1")
        self.testobj.buffs.remove_by_type(_TestModBuff)
        self.assertEqual(self.testobj.stat1, 10)
        self.testobj.buffs.clear()
        self.assertFalse(self.testobj.buffs.all)

    @patch("evennia.contrib.rpg.buffs.buff.utils.delay", new=Mock())
    def test_modgen(self):
        """test generating mods on the fly"""
        # setup
        handler: BuffHandler = self.testobj.buffs
        self.testobj.db.gentest = 5
        self.assertEqual(self.testobj.db.gentest, 5)
        tc = {"modgen": ["gentest", "add", 5, 0]}
        handler.add(StatBuff, key="gentest", to_cache=tc)
        self.assertEqual(handler.check(self.testobj.db.gentest, "gentest"), 10)
        tc = {"modgen": ["gentest", "add", 10, 0]}
        handler.add(StatBuff, key="gentest", to_cache=tc)
        self.assertEqual(handler.check(self.testobj.db.gentest, "gentest"), 15)
        self.assertEqual(
            handler.get("gentest").flavor, "This buff affects the following stats: gentest"
        )



"""

MSSP (Mud Server Status Protocol) meta information

Modify this file to specify what MUD listing sites will report about your game.
All fields are static. The number of currently active players and your game's
current uptime will be added automatically by Evennia.

You don't have to fill in everything (and most fields are not shown/used by all
crawlers anyway); leave the default if so needed. You need to reload the server
before the updated information is made available to crawlers (reloading does
not affect uptime).

After changing the values in this file, you must register your game with the
MUD website list you want to track you. The listing crawler will then regularly
connect to your server to get the latest info. No further configuration is
needed on the Evennia side.

"""

MSSPTable = {
    # Required fields
    "NAME": "Mygame",  # usually the same as SERVERNAME
    # Generic
    "CRAWL DELAY": "-1",  # limit how often crawler may update the listing. -1 for no limit
    "HOSTNAME": "",  # telnet hostname
    "PORT": ["4000"],  # telnet port - most important port should be *last* in list!
    "CODEBASE": "Evennia",
    "CONTACT": "",  # email for contacting the mud
    "CREATED": "",  # year MUD was created
    "ICON": "",  # url to icon 32x32 or larger; <32kb.
    "IP": "",  # current or new IP address
    "LANGUAGE": "",  # name of language used, e.g. English
    "LOCATION": "",  # full English name of server country
    "MINIMUM AGE": "0",  # set to 0 if not applicable
    "WEBSITE": "",  # http:// address to your game website
    # Categorisation
    "FAMILY": "Evennia",
    "GENRE": "None",  # Adult, Fantasy, Historical, Horror, Modern, None, or Science Fiction
    # Gameplay: Adventure, Educational, Hack and Slash, None,
    # Player versus Player, Player versus Environment,
    # Roleplaying, Simulation, Social or Strategy
    "GAMEPLAY": "",
    "STATUS": "Open Beta",  # Allowed: Alpha, Closed Beta, Open Beta, Live
    "GAMESYSTEM": "Custom",  # D&D, d20 System, World of Darkness, etc. Use Custom if homebrew
    # Subgenre: LASG, Medieval Fantasy, World War II, Frankenstein,
    # Cyberpunk, Dragonlance, etc. Or None if not applicable.
    "SUBGENRE": "None",
    # World
    "AREAS": "0",
    "HELPFILES": "0",
    "MOBILES": "0",
    "OBJECTS": "0",
    "ROOMS": "0",  # use 0 if room-less
    "CLASSES": "0",  # use 0 if class-less
    "LEVELS": "0",  # use 0 if level-less
    "RACES": "0",  # use 0 if race-less
    "SKILLS": "0",  # use 0 if skill-less
    # Protocols set to 1 or 0; should usually not be changed)
    "ANSI": "1",
    "GMCP": "1",
    "MSDP": "1",
    "MXP": "1",
    "SSL": "1",
    "UTF-8": "1",
    "MCCP": "1",
    "XTERM 256 COLORS": "1",
    "XTERM TRUE COLORS": "0",
    "ATCP": "0",
    "MCP": "0",
    "MSP": "0",
    "VT100": "0",
    "PUEBLO": "0",
    "ZMP": "0",
    # Commercial set to 1 or 0)
    "PAY TO PLAY": "0",
    "PAY FOR PERKS": "0",
    # Hiring  set to 1 or 0)
    "HIRING BUILDERS": "0",
    "HIRING CODERS": "0",
    # Extended variables
    # World
    "DBSIZE": "0",
    "EXITS": "0",
    "EXTRA DESCRIPTIONS": "0",
    "MUDPROGS": "0",
    "MUDTRIGS": "0",
    "RESETS": "0",
    # Game  (set to 1 or 0, or one of the given alternatives)
    "ADULT MATERIAL": "0",
    "MULTICLASSING": "0",
    "NEWBIE FRIENDLY": "0",
    "PLAYER CITIES": "0",
    "PLAYER CLANS": "0",
    "PLAYER CRAFTING": "0",
    "PLAYER GUILDS": "0",
    "EQUIPMENT SYSTEM": "None",  # "None", "Level", "Skill", "Both"
    "MULTIPLAYING": "None",  # "None", "Restricted", "Full"
    "PLAYERKILLING": "None",  # "None", "Restricted", "Full"
    "QUEST SYSTEM": "None",  # "None", "Immortal Run", "Automated", "Integrated"
    "ROLEPLAYING": "None",  # "None", "Accepted", "Encouraged", "Enforced"
    "TRAINING SYSTEM": "None",  # "None", "Level", "Skill", "Both"
    # World originality: "All Stock", "Mostly Stock", "Mostly Original", "All Original"
    "WORLD ORIGINALITY": "All Original",
}



# coding=utf-8
"""
EvForm - a way to create advanced ASCII forms

This is intended for creating advanced ASCII game forms, such as a large pretty character sheet or
info document.

The system works on the basis of a readin template that is given in a separate Python file imported
into the handler. This file contains some optional settings and a string mapping out the form. The
template has markers in it to denounce fields to fill. The markers map the absolute size of the
field and will be filled with an `evtable.EvCell` object when displaying the form.

Example of input file `testform.py`:

```python
FORMCHAR = "x"
TABLECHAR = "c"

FORM = '''
.------------------------------------------------.
|                                                |
|  Name: xxxxx1xxxxx    Player: xxxxxxx2xxxxxxx  |
|        xxxxxxxxxxx                             |
|                                                |
 >----------------------------------------------<
|                                                |
| Desc:  xxxxxxxxxxx    STR: x4x    DEX: x5x     |
|        xxxxx3xxxxx    INT: x6x    STA: x7x     |
|        xxxxxxxxxxx    LUC: x8x    MAG: x9x     |
|                                                |
 >----------------------------------------------<
|          |                                     |
| cccccccc | ccccccccccccccccccccccccccccccccccc |
| cccccccc | ccccccccccccccccccccccccccccccccccc |
| cccAcccc | ccccccccccccccccccccccccccccccccccc |
| cccccccc | ccccccccccccccccccccccccccccccccccc |
| cccccccc | cccccccccccccccccBccccccccccccccccc |
|          |                                     |
|                                           v&   |
-------------------------------------------------
'''
```

The first line of the `FORM` string is ignored if empty. The forms and table markers must mark out
complete, unbroken rectangles, each containing one embedded single-character identifier (so the
smallest element possible is a 3-character wide form). The identifier can be any character except
for the `FORM_CHAR` and `TABLE_CHAR` and some of the common ASCII-art elements, like space, `_` `|`
`*` etc (see `INVALID_FORMCHARS` in this module). Form Rectangles can have any size, but must be
separated from each other by at least one other character's width.

The form can also replace literal markers not abiding by these rules. For example, the `v&` in the
bottom right corner could be such literal marker. If a literal-mapping for 'v&' is provided, all
occurrences of this marker will be replaced. This will happen *before* any other parsing, so in
principle this could be used to inject new fields/tables into the form dynamically.  This literal
mapping does not consider width, but it will affect to total width of the form, so make sure what
you inject does not break things.  Using literal markers is the only way to inject 1 or 2-character
replacements.

Usage

```python
from evennia import EvForm, EvTable

# create a new form from the template - using the python path
form = EvForm("path.to.testform")

# alteratively, you can supply the template as a dict:

form = EvForm({"FORM": "....", "TABLECHAR": "c", "FORMCHAR": "x"})

# EvForm can also take a dictionary instead of a filepath, as long
# as the dict contains the keys FORMCHAR, TABLECHAR and FORM
# form = EvForm(form=form_dict)

# add data to each tagged form cell
form.map(cells={1: "Tom the Bouncer",
                2: "Griatch",
                3: "A sturdy fellow",
                4: 12,
                5: 10,
                6:  5,
                7: 18,
                8: 10,
                9:  3})
# create the EvTables
tableA = EvTable("HP","MV","MP",
                           table=[["**"], ["*****"], ["***"]],
                           border="incols")
tableB = EvTable("Skill", "Value", "Exp",
                           table=[["Shooting", "Herbalism", "Smithing"],
                                  [12,14,9],["550/1200", "990/1400", "205/900"]],
                           border="incols")
# map 'literal' replacents (here, a version string)
custom_mapping = {"v&", "v2"}

# add the tables to the proper ids in the form
form.map(tables={"A": tableA,
                 "B": tableB})

print(form)
```

This produces the following result:

::

    .------------------------------------------------.
    |                                                |
    |  Name: Tom the        Player: Griatch          |
    |        Bouncer                                 |
    |                                                |
     >----------------------------------------------<
    |                                                |
    | Desc:  A sturdy       STR: 12     DEX: 10      |
    |        fellow         INT: 5      STA: 18      |
    |                       LUC: 10     MAG: 3       |
    |                                                |
     >----------------------------------------------<
    |          |                                     |
    | HP|MV|MP | Skill      |Value      |Exp         |
    | ~~+~~+~~ | ~~~~~~~~~~~+~~~~~~~~~~~+~~~~~~~~~~~ |
    | **|**|** | Shooting   |12         |550/1200    |
    |   |**|*  | Herbalism  |14         |990/1400    |
    |   |* |   | Smithing   |9          |205/900     |
    |          |                                     |
    |                                           v2   |
     ------------------------------------------------

The marked forms have been replaced with EvCells of text and with EvTables. The literal marker `v&`
was replaced with `v2`.

If you change the form layout on disk, you can use `form.reload()` to re-read it from disk without
creating a new form.

If you want to update the data of an existing form, you can use `form.map()` with the changes - the
mappings will be updated, keeping the things you want. You can also update the template itself this
way, by supplying it as a dict.

Each component (except literal mappings) is restrained to the width and height specified by the
template, so it will resize to fit (or crop text if the area is too small for it). If you try to fit
a table into an area it cannot fit into (when including its borders and at least one line of text),
the form will raise an error.

----

"""

import re
from copy import copy

from evennia.utils.ansi import ANSIString
from evennia.utils.ansi import raw as ansi_raw
from evennia.utils.evtable import EvCell, EvTable
from evennia.utils.utils import all_from_module, is_iter, to_str

# non-valid form-identifying characters (which can thus be
# used as separators between forms without being detected
# as an identifier). These should be listed in regex form.
INVALID_FORMCHARS = r"\s\/\|\\\*\_\-\#\<\>\~\^\:\;\.\,"
# if there is an ansi-escape (||) we have to replace this with ||| to make sure
# to properly escape down the line
_ANSI_ESCAPE = re.compile(r"\|\|")


class EvForm:
    """
    This object is instantiated with a text file and parses
    it for rectangular form fields. It can then be fed a
    mapping so as to populate the fields with fixed-width
    EvCell or Tables.

    """

    # cell option defaults
    cell_options = {
        "pad_left": 0,
        "pad_right": 0,
        "pad_top": 0,
        "pad_bottom": 0,
        "align": "l",
        "valign": "t",
        "enforce_size": True,
    }

    # table option defaults
    table_options = {
        "pad_left": 0,
        "pad_right": 0,
        "pad_top": 0,
        "pad_bottom": 0,
        "align": "l",
        "valign": "t",
        "enforce_size": True,
    }

    def __init__(self, data=None, cells=None, tables=None, literals=None, **kwargs):
        """
        Initiate the form

        Keyword Args:
            data (str or dict): Path to template file or a dict with
                "formchar", "tablechar" and "form" keys (not case sensitive, so FORM etc
                also works, to stay compatible with the in-file names). While "form/FORM"
                is required, if FORMCHAR/TABLECHAR are not given, they will default to
                'x' and 'c' respectively.
            cells (dict): A dictionary mapping  `{id: str}`
            tables (dict): A dictionary mapping  `{id: EvTable}`.
            literals (dict): A dictionary mapping `{id: str}`. Will be replaced
                after width of form is calculated, but before cells/tables are mapped.
                All occurrences of the identifier on the form will be replaced. Note
                that there is no length-restriction on the remap, you are responsible
                for not breaking the form.

        Notes:
            Other kwargs are fed as options to the EvCells and EvTables
            (see `evtable.EvCell` and `evtable.EvTable` for more info).

        """
        self.indata = data  # storing here so we can reload later in case of a filename
        self.options = self._parse_inkwargs(**kwargs)

        self.cells_mapping = (
            dict((to_str(key), value) for key, value in cells.items()) if cells else {}
        )
        self.tables_mapping = (
            dict((to_str(key), value) for key, value in tables.items()) if tables else {}
        )
        self.literals_mapping = (
            dict((to_str(key), to_str(value)) for key, value in literals.items())
            if literals
            else {}
        )

        # work arrays
        self.literal_form = ""
        self.mapping = {}
        self.matrix = []
        self.form = []

        # will parse and build the form
        self.reload()

    def _parse_indata(self):
        """
        Parse and validate the `self.indata` property. We do this in order to be able to
        re-load the evform module if indata is a filename and catch any on-file changes.

        Returns:
            dict: The data dict parsed/generated from the in-data.

        """
        data = self.indata

        default_formchar = "x"
        default_tablechar = "c"

        if isinstance(data, str):
            # a module path - read all variables from it
            data = all_from_module(data)

        if isinstance(data, dict):
            data = {
                "form": str(data.get("form", data.get("FORM", None))),
                "formchar": str(data.get("formchar", data.get("FORMCHAR", default_formchar))),
                "tablechar": str(data.get("tablechar", data.get("TABLECHAR", default_tablechar))),
            }
        else:
            raise RuntimeError(f"EvForm invalid input: {data}.")

        if not data or data["form"] is None:
            raise RuntimeError("Evform data must specify a valid 'form' or 'FORM'.")

        # handle empty or multi-character form/tablechars (not supported)
        data["formchar"] = data["formchar"][0] if data["formchar"] else default_formchar
        data["tablechar"] = data["tablechar"][0] if data["tablechar"] else default_tablechar
        if re.match(rf"[{INVALID_FORMCHARS}]", data["formchar"]):
            raise RuntimeError(f"Invalid formchar: {data['formchar']}")
        if re.match(rf"[{INVALID_FORMCHARS}]", data["tablechar"]):
            raise RuntimeError(f"Invalid tablechar: {data['tablechar']}")

        return data

    def _parse_inkwargs(self, **kwargs):
        """
        Validate incoming kwargs that will be passed on to become cell/table options.

        Keyword Args:
            any: Kwargs to process.

        Returns:
            dict: A validated/cleaned kwarg to use for options.

        """
        if "filename" in kwargs:
            raise DeprecationWarning(
                "EvForm's 'filename' kwarg was renamed to 'data' and can now accept both "
                "a python path and a dict with 'FORMCHAR', 'TABLECHAR' and 'FORM' keys."
            )
        if "form" in kwargs:
            raise DeprecationWarning(
                "EvForms's 'form' kwarg was renamed to 'data' and can now accept both "
                "a python path and a dict detailing the form."
            )

        # clean cell kwarg options (these cannot be overridden on the cell but must be controlled
        # by the evform itself)
        kwargs.pop("enforce_size", None)
        kwargs.pop("width", None)
        kwargs.pop("height", None)

        return kwargs

    def _do_literal_mapping(self):
        """
        Do literal replacement in the EvForm.

        """
        literal_form = copy(self.data["form"])

        for key, repl in self.literals_mapping.items():
            literal_form = literal_form.replace(key, repl)
        return literal_form

    def _parse_to_matrix(self):
        """
        Forces all lines to be as long as the longest line, filling with whitespace.

        Args:
            lines (list): list of `ANSIString`s

        Returns:
            (list): list of `ANSIString`s of
            same length as the longest input line

        """
        matrix = EvForm._to_ansi(self.literal_form.split("\n"))

        maxl = max(len(line) for line in matrix)
        matrix = [line + " " * (maxl - len(line)) for line in matrix]
        if matrix and not matrix[0].strip():
            # the first line is normally empty, we strip it.
            matrix = matrix[1:]
        return matrix

    @staticmethod
    def _to_ansi(obj, regexable=False):
        "convert anything to ANSIString"

        if isinstance(obj, ANSIString):
            return obj
        elif isinstance(obj, str):
            # since ansi will be parsed twice (here and in the normal ansi send), we have to
            # escape ansi twice.
            obj = ansi_raw(obj)

        if isinstance(obj, dict):
            return dict(
                (key, EvForm._to_ansi(value, regexable=regexable)) for key, value in obj.items()
            )
        # regular _to_ansi (from EvTable)
        elif is_iter(obj):
            return [EvForm._to_ansi(o) for o in obj]
        else:
            return ANSIString(obj, regexable=regexable)

    def _rectangles_to_mapping(self):
        """
        Parse a form for rectangular formfields identified by formchar/tablechar enclosing an
        identifier.

        """
        formchar = self.data["formchar"]
        tablechar = self.data["tablechar"]
        matrix = self.matrix

        cell_options = copy(self.cell_options)
        cell_options.update(self.options)

        table_options = copy(self.table_options)
        table_options.update(self.options)

        nmatrix = len(matrix)

        mapping = {}

        def _get_rectangles(char):
            """Find all identified rectangles marked with given char"""
            rects = []
            coords = {}
            regex = re.compile(rf"{char}+([^{INVALID_FORMCHARS}{char}]+){char}+")

            # find the start/width of rectangles for each line
            for iy, line in enumerate(EvForm._to_ansi(matrix, regexable=True)):
                ix0 = 0
                while True:
                    match = regex.search(line, ix0)
                    if match:
                        # get the width of the rectangle directly from the match
                        coords[match.group(1)] = [iy, match.start(), match.end()]
                        ix0 = match.end()
                    else:
                        break

            for key, (iy, leftix, rightix) in coords.items():
                # scan up to find top of rectangle
                dy_up = 0
                if iy > 0:
                    for i in range(1, iy):
                        if all(matrix[iy - i][ix] == char for ix in range(leftix, rightix)):
                            dy_up += 1
                        else:
                            break
                # find bottom edge of rectangle
                dy_down = 0
                if iy < nmatrix - 1:
                    for i in range(1, nmatrix - iy - 1):
                        if all(matrix[iy + i][ix] == char for ix in range(leftix, rightix)):
                            dy_down += 1
                        else:
                            break

                #  we have our rectangle. Calculate size
                iyup = iy - dy_up
                iydown = iy + dy_down
                width = rightix - leftix
                height = abs(iyup - iydown) + 1

                # store (key, y, x, width, height) of triangle
                rects.append((key, iyup, leftix, width, height))

            return rects

        # Map EvCells into form rectangles
        for key, y, x, width, height in _get_rectangles(formchar):
            # get data to populate cell
            data = self.cells_mapping.get(key, "")
            if isinstance(data, EvCell):
                # mapping already provides the cell. We need to override some
                # of the cell's options to make it work in the evform rectangle.
                # We retain the align/valign since this may be interesting to
                # play with within the rectangle.
                cell = data
                custom_align = cell.align
                custom_valign = cell.valign
                cell.reformat(
                    width=width,
                    height=height,
                    **{**cell_options, **{"align": custom_align, "valign": custom_valign}},
                )
            else:
                # generating cell on the fly
                cell = EvCell(data, width=width, height=height, **cell_options)

            mapping[key] = (y, x, width, height, cell)

        # Map EvTables into form rectangles
        for key, y, x, width, height in _get_rectangles(tablechar):
            # get EvTable from mapping
            table = self.tables_mapping.get(key, None)

            if table:
                table.reformat(width=width, height=height, **table_options)
            else:
                table = EvTable(width=width, height=height, **table_options)

            mapping[key] = (y, x, width, height, table)

        return mapping

    def _build_form(self):
        """
        Insert cell/table contents into form at given locations to create
        the final result.

        """
        form = copy(self.matrix)
        mapping = self.mapping

        for key, (y, x, width, height, cell_or_table) in mapping.items():
            # rect is a list of <height> lines, each <width> wide
            rect = cell_or_table.get()
            for il, rectline in enumerate(rect):
                formline = form[y + il]
                # insert new content, replacing old
                form[y + il] = formline[:x] + rectline + formline[x + width :]

        return form

    def reload(self):
        """
        Creates the form from a filename or data structure.

        Args:
            data (str or dict): Can be used to update an existing form using
                the same cells/tables provided on initialization or using `.map()`.

        Notes:
            Kwargs are passed through to Cel creation.

        """
        self.data = self._parse_indata()

        # Map any literals into the string
        self.literal_form = self._do_literal_mapping()
        # Create raw form matrix, indexable with (y, x) coords
        self.matrix = self._parse_to_matrix()
        # parse and identify all rectangles in the form
        self.mapping = self._rectangles_to_mapping()
        # combine mapping with form template into a final result
        self.form = self._build_form()

    def map(self, cells=None, tables=None, data=None, literals=None, **kwargs):
        """
        Add mapping for form. This allows for updating an existing
        evform.

        Args:
            cells (dict): A dictionary of {identifier:celltext}. These
                will be appended to the existing mappings.
            tables (dict): A dictionary of {identifier:table}. Will
                be appended to the existing mapping.
            data (str or dict): A path to a evform module or a dict with
                the needed "FORM", "TABLE/FORMCHAR" keys. Will replace
                the originally initialized form.
            literals

        Keyword Args:
            These will be appended to the existing cell/table options.

        Notes:
            kwargs will be forwarded to tables/cells. See
            `evtable.EvCell` and `evtable.EvTable` for info.

        """
        if data:
            # storing so ._parse_indata will find it during reload
            self.indata = data

        new_cells = dict((to_str(key), value) for key, value in cells.items()) if cells else {}
        self.cells_mapping.update(new_cells)
        new_tables = dict((to_str(key), value) for key, value in tables.items()) if tables else {}
        self.tables_mapping.update(new_tables)
        new_literals = (
            dict((to_str(key), to_str(value)) for key, value in literals.items())
            if literals
            else {}
        )
        self.literals_mapping.update(new_literals)

        self.options.update(self._parse_inkwargs(**kwargs))

        # parse and build the form
        self.reload()

    def __str__(self):
        "Prints the form"
        return str(ANSIString("\n").join([line for line in self.form]))



"""
Tests of ingame_map_display.

"""


from typeclasses import exits, rooms

from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils.create import create_object

from . import ingame_map_display


class TestIngameMap(BaseEvenniaCommandTest):
    """
    Test the ingame map display by building two rooms and checking their connections are found

    Expected output:
    [ ]--[ ]
    """

    def setUp(self):
        super().setUp()
        self.west_room = create_object(rooms.Room, key="Room 1")
        self.east_room = create_object(rooms.Room, key="Room 2")
        create_object(
            exits.Exit,
            key="east",
            aliases=["e"],
            location=self.west_room,
            destination=self.east_room,
        )
        create_object(
            exits.Exit,
            key="west",
            aliases=["w"],
            location=self.east_room,
            destination=self.west_room,
        )

    def test_west_room_map_room(self):
        self.char1.location = self.west_room
        map_here = ingame_map_display.Map(self.char1).show_map()
        self.assertEqual(map_here.strip(), "[|n|[x|co|n]|n--[|n ]|n")

    def test_east_room_map_room(self):
        self.char1.location = self.east_room
        map_here = ingame_map_display.Map(self.char1).show_map()
        self.assertEqual(map_here.strip(), "[|n ]|n--[|n|[x|co|n]|n")



"""
Typeclass for Account objects.

Note that this object is primarily intended to
store OOC information, not game info! This
object represents the actual user (not their
character) and has NO actual presence in the
game world (this is handled by the associated
character object, so you should customize that
instead for most things).

"""
import re
import time
import typing
from random import getrandbits

from django.conf import settings
from django.contrib.auth import authenticate, password_validation
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _

import evennia
from evennia.accounts.manager import AccountManager
from evennia.accounts.models import AccountDB
from evennia.commands.cmdsethandler import CmdSetHandler
from evennia.comms.models import ChannelDB
from evennia.objects.models import ObjectDB
from evennia.scripts.scripthandler import ScriptHandler
from evennia.server.models import ServerConfig
from evennia.server.signals import (
    SIGNAL_ACCOUNT_POST_CREATE,
    SIGNAL_ACCOUNT_POST_LOGIN_FAIL,
    SIGNAL_OBJECT_POST_PUPPET,
    SIGNAL_OBJECT_POST_UNPUPPET,
)
from evennia.server.throttle import Throttle
from evennia.typeclasses.attributes import ModelAttributeBackend, NickHandler
from evennia.typeclasses.models import TypeclassBase
from evennia.utils import class_from_module, create, logger
from evennia.utils.optionhandler import OptionHandler
from evennia.utils.utils import (
    is_iter,
    lazy_property,
    make_iter,
    to_str,
    variable_from_module,
)

__all__ = ("DefaultAccount", "DefaultGuest")

_AT_SEARCH_RESULT = variable_from_module(*settings.SEARCH_AT_RESULT.rsplit(".", 1))
_MULTISESSION_MODE = settings.MULTISESSION_MODE
_AUTO_CREATE_CHARACTER_WITH_ACCOUNT = settings.AUTO_CREATE_CHARACTER_WITH_ACCOUNT
_AUTO_PUPPET_ON_LOGIN = settings.AUTO_PUPPET_ON_LOGIN
_MAX_NR_SIMULTANEOUS_PUPPETS = settings.MAX_NR_SIMULTANEOUS_PUPPETS
_MAX_NR_CHARACTERS = settings.MAX_NR_CHARACTERS
_CMDSET_ACCOUNT = settings.CMDSET_ACCOUNT
_MUDINFO_CHANNEL = None
_CONNECT_CHANNEL = None
_CMDHANDLER = None


# Create throttles for too many account-creations and login attempts
CREATION_THROTTLE = Throttle(
    name="creation",
    limit=settings.CREATION_THROTTLE_LIMIT,
    timeout=settings.CREATION_THROTTLE_TIMEOUT,
)
LOGIN_THROTTLE = Throttle(
    name="login", limit=settings.LOGIN_THROTTLE_LIMIT, timeout=settings.LOGIN_THROTTLE_TIMEOUT
)


class AccountSessionHandler(object):
    """
    Manages the session(s) attached to an account.

    """

    def __init__(self, account):
        """
        Initializes the handler.

        Args:
            account (Account): The Account on which this handler is defined.

        """
        self.account = account

    def get(self, sessid=None):
        """
        Get the sessions linked to this object.

        Args:
            sessid (int, optional): Specify a given session by
                session id.

        Returns:
            sessions (list): A list of Session objects. If `sessid`
                is given, this is a list with one (or zero) elements.

        """
        if sessid:
            return make_iter(evennia.SESSION_HANDLER.session_from_account(self.account, sessid))
        else:
            return evennia.SESSION_HANDLER.sessions_from_account(self.account)

    def all(self):
        """
        Alias to get(), returning all sessions.

        Returns:
            sessions (list): All sessions.

        """
        return self.get()

    def count(self):
        """
        Get amount of sessions connected.

        Returns:
            sesslen (int): Number of sessions handled.

        """
        return len(self.get())


class CharactersHandler:
    """
    A simple Handler that lives on DefaultAccount as .characters via @lazy_property used to
    wrap access to .db._playable_characters.
    """

    def __init__(self, owner: "DefaultAccount"):
        """
        Create the CharactersHandler.

        Args:
            owner: The Account that owns this handler.
        """
        self.owner = owner
        self._ensure_playable_characters()
        self._clean()

    def _ensure_playable_characters(self):
        if self.owner.db._playable_characters is None:
            self.owner.db._playable_characters = []

    def _clean(self):
        # Remove all instances of None from the list.
        self.owner.db._playable_characters = [x for x in self.owner.db._playable_characters if x]

    def add(self, character: "DefaultCharacter"):
        """
        Add a character to this account's list of playable characters.

        Args:
            character (DefaultCharacter): The character to add.
        """
        self._clean()
        if character not in self.owner.db._playable_characters:
            self.owner.db._playable_characters.append(character)
            self.owner.at_post_add_character(character)

    def remove(self, character: "DefaultCharacter"):
        """
        Remove a character from this account's list of playable characters.

        Args:
            character (DefaultCharacter): The character to remove.
        """
        self._clean()
        if character in self.owner.db._playable_characters:
            self.owner.db._playable_characters.remove(character)
            self.owner.at_post_remove_character(character)

    def all(self) -> list["DefaultCharacter"]:
        """
        Get all playable characters.

        Returns:
            list[DefaultCharacter]: All playable characters.
        """
        self._clean()
        return list(self.owner.db._playable_characters)

    def count(self) -> int:
        """
        Get the number of playable characters.

        Returns:
            int: The number of playable characters.
        """
        return len(self.all())

    __len__ = count

    def __iter__(self):
        return iter(self.all())


class DefaultAccount(AccountDB, metaclass=TypeclassBase):
    """
    This is the base Typeclass for all Accounts. Accounts represent
    the person playing the game and tracks account info, password
    etc. They are OOC entities without presence in-game. An Account
    can connect to a Character Object in order to "enter" the
    game.

    Account Typeclass API:

    * Available properties (only available on initiated typeclass objects)

     - key (string) - name of account
     - name (string)- wrapper for user.username
     - aliases (list of strings) - aliases to the object. Will be saved to
            database as AliasDB entries but returned as strings.
     - dbref (int, read-only) - unique #id-number. Also "id" can be used.
     - date_created (string) - time stamp of object creation
     - permissions (list of strings) - list of permission strings
     - user (User, read-only) - django User authorization object
     - obj (Object) - game object controlled by account. 'character' can also
                     be used.
     - sessions (list of Sessions) - sessions connected to this account
     - is_superuser (bool, read-only) - if the connected user is a superuser

    * Handlers

     - locks - lock-handler: use locks.add() to add new lock strings
     - db - attribute-handler: store/retrieve database attributes on this
                              self.db.myattr=val, val=self.db.myattr
     - ndb - non-persistent attribute handler: same as db but does not
                                  create a database entry when storing data
     - scripts - script-handler. Add new scripts to object with scripts.add()
     - cmdset - cmdset-handler. Use cmdset.add() to add new cmdsets to object
     - nicks - nick-handler. New nicks with nicks.add().

    * Helper methods

     - msg(text=None, from_obj=None, session=None, options=None, **kwargs)
     - execute_cmd(raw_string)
     - search(ostring, global_search=False, attribute_name=None,
                      use_nicks=False, location=None,
                      ignore_errors=False, account=False)
     - is_typeclass(typeclass, exact=False)
     - swap_typeclass(new_typeclass, clean_attributes=False, no_default=True)
     - access(accessing_obj, access_type='read', default=False, no_superuser_bypass=False)
     - check_permstring(permstring)

    * Hook methods

     basetype_setup()
     at_account_creation()

     > note that the following hooks are also found on Objects and are
       usually handled on the character level:

     - at_init()
     - at_access()
     - at_cmdset_get(**kwargs)
     - at_first_login()
     - at_post_login(session=None)
     - at_disconnect()
     - at_message_receive()
     - at_message_send()
     - at_server_reload()
     - at_server_shutdown()

    """

    # Determines which order command sets begin to be assembled from.
    # Accounts are usually second.
    cmdset_provider_order = 50
    cmdset_provider_error_order = 0
    cmdset_provider_type = "account"

    objects = AccountManager()

    # Used by account.create_character() to choose default typeclass for characters.
    default_character_typeclass = settings.BASE_CHARACTER_TYPECLASS

    lockstring = (
        "examine:perm(Admin);edit:perm(Admin);"
        "delete:perm(Admin);boot:perm(Admin);msg:all();"
        "noidletimeout:perm(Builder) or perm(noidletimeout)"
    )

    # properties
    @lazy_property
    def cmdset(self):
        return CmdSetHandler(self, True)

    @lazy_property
    def scripts(self):
        return ScriptHandler(self)

    @lazy_property
    def nicks(self):
        return NickHandler(self, ModelAttributeBackend)

    @lazy_property
    def sessions(self):
        return AccountSessionHandler(self)

    @lazy_property
    def options(self):
        return OptionHandler(
            self,
            options_dict=settings.OPTIONS_ACCOUNT_DEFAULT,
            savefunc=self.attributes.add,
            loadfunc=self.attributes.get,
            save_kwargs={"category": "option"},
            load_kwargs={"category": "option"},
        )

    @lazy_property
    def characters(self):
        return CharactersHandler(self)

    def get_cmdset_providers(self) -> dict[str, "CmdSetProvider"]:
        """
        Overrideable method which returns a dictionary of every kind of object which
        has a cmdsethandler linked to this Account, and should participate in cmdset
        merging.

        Accounts have no way of being aware of anything besides themselves, unfortunately.

        Returns:
            dict[str, CmdSetProvider]: The CmdSetProviders linked to this Object.
        """
        return {"account": self}

    def at_post_add_character(self, character: "DefaultCharacter"):
        """
        Called after a character is added to this account's list of playable characters.

        Use it to easily implement custom logic when a character is added to an account.

        Args:
            character (DefaultCharacter): The character that was added.
        """
        pass

    def at_post_remove_character(self, character: "DefaultCharacter"):
        """
        Called after a character is removed from this account's list of playable characters.

        Use it to easily implement custom logic when a character is removed from an account.

        Args:
            character (DefaultCharacter): The character that was removed.
        """
        pass

    def uses_screenreader(self, session=None):
        """
        Shortcut to determine if a session uses a screenreader. If no session given,
        will return true if any of the sessions use a screenreader.

        Args:
            session (Session, optional): The session to check for screen reader.

        """
        if session:
            return bool(session.protocol_flags.get("SCREENREADER", False))
        else:
            return any(
                session.protocol_flags.get("SCREENREADER") for session in self.sessions.all()
            )

    def get_display_name(self, looker, **kwargs):
        """
        This is used by channels and other OOC communications methods to give a
        custom display of this account's input.

        Args:
            looker (Account): The one that will see this name.
            **kwargs: Unused by default, can be used to pass game-specific data.

        Returns:
            str: The name, possibly modified.

        """
        return f"|c{self.key}|n"

    # session-related methods

    def disconnect_session_from_account(self, session, reason=None):
        """
        Access method for disconnecting a given session from the
        account (connection happens automatically in the
        sessionhandler)

        Args:
            session (Session): Session to disconnect.
            reason (str, optional): Eventual reason for the disconnect.

        """
        evennia.SESSION_HANDLER.disconnect(session, reason)

    # puppeting operations

    def puppet_object(self, session, obj):
        """
        Use the given session to control (puppet) the given object (usually
        a Character type).

        Args:
            session (Session): session to use for puppeting
            obj (Object): the object to start puppeting

        Raises:
            RuntimeError: If puppeting is not possible, the
                `exception.msg` will contain the reason.


        """
        # safety checks
        if not obj:
            raise RuntimeError("Object not found")
        if not session:
            raise RuntimeError("Session not found")
        if self.get_puppet(session) == obj:
            # already puppeting this object
            self.msg("You are already puppeting this object.")
            return
        if not obj.access(self, "puppet"):
            # no access
            self.msg(f"You don't have permission to puppet '{obj.key}'.")
            return
        if obj.account:
            # object already puppeted
            if obj.account == self:
                if obj.sessions.count():
                    # we may take over another of our sessions
                    # output messages to the affected sessions
                    if _MULTISESSION_MODE in (1, 3):
                        txt1 = f"Sharing |c{obj.name}|n with another of your sessions."
                        txt2 = f"|c{obj.name}|n|G is now shared from another of your sessions.|n"
                        self.msg(txt1, session=session)
                        self.msg(txt2, session=obj.sessions.all())
                    else:
                        txt1 = f"Taking over |c{obj.name}|n from another of your sessions."
                        txt2 = f"|c{obj.name}|n|R is now acted from another of your sessions.|n"
                        self.msg(txt1, session=session)
                        self.msg(txt2, session=obj.sessions.all())
                        self.unpuppet_object(obj.sessions.get())
            elif obj.account.is_connected:
                # controlled by another account
                self.msg(_("|c{key}|R is already puppeted by another Account.").format(key=obj.key))
                return

        if session.puppet:
            # cleanly unpuppet eventual previous object puppeted by this session
            self.unpuppet_object(session)
        # if we get to this point the character is ready to puppet or it
        # was left with a lingering account/session reference from an unclean
        # server kill or similar

        # check so we are not puppeting too much already
        if _MAX_NR_SIMULTANEOUS_PUPPETS is not None:
            already_puppeted = self.get_all_puppets()
            if (
                not self.is_superuser
                and not self.check_permstring("Developer")
                and obj not in already_puppeted
                and len(self.get_all_puppets()) >= _MAX_NR_SIMULTANEOUS_PUPPETS
            ):
                self.msg(
                    _(f"You cannot control any more puppets (max {_MAX_NR_SIMULTANEOUS_PUPPETS})")
                )
                return

        # do the puppeting
        obj.at_pre_puppet(self, session=session)
        # used to track in case of crash so we can clean up later
        obj.tags.add("puppeted", category="account")

        # do the connection
        obj.sessions.add(session)
        obj.account = self
        session.puid = obj.id
        session.puppet = obj

        # re-cache locks to make sure superuser bypass is updated
        obj.locks.cache_lock_bypass(obj)
        # final hook
        obj.at_post_puppet()
        SIGNAL_OBJECT_POST_PUPPET.send(sender=obj, account=self, session=session)

    def unpuppet_object(self, session):
        """
        Disengage control over an object.

        Args:
            session (Session or list): The session or a list of
                sessions to disengage from their puppets.

        Raises:
            RuntimeError With message about error.

        """
        for session in make_iter(session):
            obj = session.puppet
            if obj:
                # do the disconnect, but only if we are the last session to puppet
                obj.at_pre_unpuppet()
                obj.sessions.remove(session)
                if not obj.sessions.count():
                    del obj.account
                obj.at_post_unpuppet(self, session=session)
                obj.tags.remove("puppeted", category="account")
                SIGNAL_OBJECT_POST_UNPUPPET.send(sender=obj, session=session, account=self)
            # Just to be sure we're always clear.
            session.puppet = None
            session.puid = None

    def unpuppet_all(self):
        """
        Disconnect all puppets. This is called by server before a
        reset/shutdown.
        """
        self.unpuppet_object(self.sessions.all())

    def get_puppet(self, session):
        """
        Get an object puppeted by this session through this account. This is
        the main method for retrieving the puppeted object from the
        account's end.

        Args:
            session (Session): Find puppeted object based on this session

        Returns:
            puppet (Object): The matching puppeted object, if any.

        """
        return session.puppet if session else None

    def get_all_puppets(self):
        """
        Get all currently puppeted objects.

        Returns:
            puppets (list): All puppeted objects currently controlled
                by this Account.

        """
        return list(set(session.puppet for session in self.sessions.all() if session.puppet))

    def __get_single_puppet(self):
        """
        This is a legacy convenience link for use with `MULTISESSION_MODE`.

        Returns:
            puppets (Object or list): Users of `MULTISESSION_MODE` 0 or 1 will
                always get the first puppet back. Users of higher `MULTISESSION_MODE`s will
                get a list of all puppeted objects.

        """
        puppets = self.get_all_puppets()
        if _MULTISESSION_MODE in (0, 1):
            return puppets and puppets[0] or None
        return puppets

    character = property(__get_single_puppet)
    puppet = property(__get_single_puppet)

    # utility methods
    @classmethod
    def is_banned(cls, **kwargs):
        """
        Checks if a given username or IP is banned.

        Keyword Args:
            ip (str, optional): IP address.
            username (str, optional): Username.

        Returns:
            is_banned (bool): Whether either is banned or not.

        """

        ip = kwargs.get("ip", "")
        if isinstance(ip, (tuple, list)):
            ip = ip[0]
        ip = ip.strip()
        username = kwargs.get("username", "").lower().strip()

        # Check IP and/or name bans
        bans = ServerConfig.objects.conf("server_bans")
        if bans and (
            any(tup[0] == username for tup in bans if username)
            or any(tup[2].match(ip) for tup in bans if ip and tup[2])
        ):
            return True

        return False

    @classmethod
    def get_username_validators(
        cls, validator_config=getattr(settings, "AUTH_USERNAME_VALIDATORS", [])
    ):
        """
        Retrieves and instantiates validators for usernames.

        Args:
            validator_config (list): List of dicts comprising the battery of
                validators to apply to a username.

        Returns:
            validators (list): List of instantiated Validator objects.
        """
        objs = []
        for validator in validator_config:
            try:
                klass = import_string(validator["NAME"])
            except ImportError:
                msg = (
                    f"The module in NAME could not be imported: {validator['NAME']}. "
                    "Check your AUTH_USERNAME_VALIDATORS setting."
                )
                raise ImproperlyConfigured(msg)
            objs.append(klass(**validator.get("OPTIONS", {})))
        return objs

    @classmethod
    def authenticate(cls, username, password, ip="", **kwargs):
        """
        Checks the given username/password against the database to see if the
        credentials are valid.

        Note that this simply checks credentials and returns a valid reference
        to the user-- it does not log them in!

        To finish the job:
        After calling this from a Command, associate the account with a Session:
        - session.sessionhandler.login(session, account)

        ...or after calling this from a View, associate it with an HttpRequest:
        - django.contrib.auth.login(account, request)

        Args:
            username (str): Username of account
            password (str): Password of account
            ip (str, optional): IP address of client

        Keyword Args:
            session (Session, optional): Session requesting authentication

        Returns:
            account (DefaultAccount, None): Account whose credentials were
                provided if not banned.
            errors (list): Error messages of any failures.

        """
        errors = []
        if ip:
            ip = str(ip)

        # See if authentication is currently being throttled
        if ip and LOGIN_THROTTLE.check(ip):
            errors.append(_("Too many login failures; please try again in a few minutes."))

            # With throttle active, do not log continued hits-- it is a
            # waste of storage and can be abused to make your logs harder to
            # read and/or fill up your disk.
            return None, errors

        # Check IP and/or name bans
        banned = cls.is_banned(username=username, ip=ip)
        if banned:
            # this is a banned IP or name!
            errors.append(
                _(
                    "|rYou have been banned and cannot continue from here."
                    "\nIf you feel this ban is in error, please email an admin.|x"
                )
            )
            logger.log_sec(f"Authentication Denied (Banned): {username} (IP: {ip}).")
            LOGIN_THROTTLE.update(ip, "Too many sightings of banned artifact.")
            return None, errors

        # Authenticate and get Account object
        account = authenticate(username=username, password=password)
        if not account:
            # User-facing message
            errors.append(_("Username and/or password is incorrect."))

            # Log auth failures while throttle is inactive
            logger.log_sec(f"Authentication Failure: {username} (IP: {ip}).")

            # Update throttle
            if ip:
                LOGIN_THROTTLE.update(ip, _("Too many authentication failures."))

            # Try to call post-failure hook
            session = kwargs.get("session", None)
            if session:
                account = AccountDB.objects.get_account_from_name(username)
                if account:
                    SIGNAL_ACCOUNT_POST_LOGIN_FAIL.send(sender=account, session=session)
                    account.at_failed_login(session)

            return None, errors

        # Account successfully authenticated
        logger.log_sec(f"Authentication Success: {account} (IP: {ip}).")
        return account, errors

    @classmethod
    def normalize_username(cls, username):
        """
        Django: Applies NFKC Unicode normalization to usernames so that visually
        identical characters with different Unicode code points are considered
        identical.

        (This deals with the Turkish "i" problem and similar
        annoyances. Only relevant if you go out of your way to allow Unicode
        usernames though-- Evennia accepts ASCII by default.)

        In this case we're simply piggybacking on this feature to apply
        additional normalization per Evennia's standards.
        """
        username = super(DefaultAccount, cls).normalize_username(username)

        # strip excessive spaces in accountname
        username = re.sub(r"\s+", " ", username).strip()

        return username

    @classmethod
    def validate_username(cls, username):
        """
        Checks the given username against the username validator associated with
        Account objects, and also checks the database to make sure it is unique.

        Args:
            username (str): Username to validate

        Returns:
            valid (bool): Whether or not the password passed validation
            errors (list): Error messages of any failures

        """
        valid = []
        errors = []

        # Make sure we're at least using the default validator
        validators = cls.get_username_validators()
        if not validators:
            validators = [cls.username_validator]

        # Try username against all enabled validators
        for validator in validators:
            try:
                valid.append(not validator(username))
            except ValidationError as e:
                valid.append(False)
                errors.extend(e.messages)

        # Disqualify if any check failed
        if False in valid:
            valid = False
        else:
            valid = True

        return valid, errors

    @classmethod
    def validate_password(cls, password, account=None):
        """
        Checks the given password against the list of Django validators enabled
        in the server.conf file.

        Args:
            password (str): Password to validate

        Keyword Args:
            account (DefaultAccount, optional): Account object to validate the
                password for. Optional, but Django includes some validators to
                do things like making sure users aren't setting passwords to the
                same value as their username. If left blank, these user-specific
                checks are skipped.

        Returns:
            valid (bool): Whether or not the password passed validation
            error (ValidationError, None): Any validation error(s) raised. Multiple
                errors can be nested within a single object.

        """
        valid = False
        error = None

        # Validation returns None on success; invert it and return a more sensible bool
        try:
            valid = not password_validation.validate_password(password, user=account)
        except ValidationError as e:
            error = e

        return valid, error

    def set_password(self, password, **kwargs):
        """
        Applies the given password to the account. Logs and triggers the `at_password_change` hook.

        Args:
            password (str): Password to set.

        Notes:
            This is called by Django also when logging in; it should not be mixed up with
            validation, since that would mean old passwords in the database (pre validation checks)
            could get invalidated.

        """
        super().set_password(password)
        logger.log_sec(f"Password successfully changed for {self}.")
        self.at_password_change()

    def get_character_slots(self) -> typing.Optional[int]:
        """
        Returns the number of character slots this account has, or
        None if there are no limits.

        By default, that's settings.MAX_NR_CHARACTERS but this makes it easy to override.
        Maybe for your game, players can be rewarded with more slots, somehow.

        Returns:
            int (optional): The number of character slots this account has, or None
                if there are no limits.
        """
        return settings.MAX_NR_CHARACTERS

    def get_available_character_slots(self) -> typing.Optional[int]:
        """
        Returns the number of character slots this account has available, or None if
        there are no limits.

        Returns:
            int (optional): The number of open character slots this account has, or None
                if there are no limits.
        """
        if (slots := self.get_character_slots()) is None:
            return None
        return max(0, slots - len(self.characters))

    def check_available_slots(self, **kwargs) -> typing.Optional[str]:
        """
        Helper method used to determine if an account can create additional characters using
        the character slot system.

        Returns:
            str (optional): An error message regarding the status of slots. If present, this
               will halt character creation. If not, character creation can proceed.
        """
        if (slots := self.get_available_character_slots()) is not None:
            if slots <= 0:
                if not (self.is_superuser or self.check_permstring("Developer")):
                    plural = "" if (max_slots := self.get_character_slots()) == 1 else "s"
                    return f"You may only have a maximum of {max_slots} character{plural}."

    def create_character(self, *args, **kwargs):
        """
        Create a character linked to this account.

        Args:
            key (str, optional): If not given, use the same name as the account.
            typeclass (str, optional): Typeclass to use for this character. If
                not given, use self.default_character_class.
            permissions (list, optional): If not given, use the account's permissions.
            ip (str, optional): The client IP creating this character. Will fall back to the
                one stored for the account if not given.
            kwargs (any): Other kwargs will be used in the create_call.
        Returns:
            Object: A new character of the `character_typeclass` type. None on an error.
            list or None: A list of errors, or None.

        """
        # check character slot usage.
        if slot_check := self.check_available_slots():
            return None, [slot_check]

        # parse inputs
        character_key = kwargs.pop("key", self.key)
        character_ip = kwargs.pop("ip", self.db.creator_ip)
        character_permissions = kwargs.pop("permissions", self.permissions)

        # Load the appropriate Character class
        character_typeclass = kwargs.pop("typeclass", self.default_character_typeclass)
        Character = class_from_module(character_typeclass)

        if "location" not in kwargs:
            kwargs["location"] = ObjectDB.objects.get_id(settings.START_LOCATION)

        # Create the character
        character, errs = Character.create(
            character_key,
            self,
            ip=character_ip,
            typeclass=character_typeclass,
            permissions=character_permissions,
            **kwargs,
        )
        if character:
            self.at_post_create_character(character, ip=character_ip)

        return character, errs

    def at_post_create_character(self, character, **kwargs):
        """
        An overloadable hook method that allows for further customization of newly created characters.
        """
        if character not in self.characters:
            self.characters.add(character)

        # We need to set this to have @ic auto-connect to this character
        if len(self.characters) == 1:
            self.db._last_puppet = character

        character.locks.add(
            f"puppet:id({character.id}) or pid({self.id}) or perm(Developer) or"
            f" pperm(Developer);delete:id({self.id}) or perm(Admin)"
        )

        logger.log_sec(
            f"Character Created: {character} (Caller: {self}, IP: {kwargs.get('ip', None)})."
        )

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Creates an Account (or Account/Character pair for MULTISESSION_MODE<2)
        with default (or overridden) permissions and having joined them to the
        appropriate default channels.

        Keyword Args:
            username (str): Username of Account owner
            password (str): Password of Account owner
            email (str, optional): Email address of Account owner
            ip (str, optional): IP address of requesting connection
            guest (bool, optional): Whether or not this is to be a Guest account

            permissions (str, optional): Default permissions for the Account
            typeclass (str, optional): Typeclass to use for new Account
            character_typeclass (str, optional): Typeclass to use for new char
                when applicable.

        Returns:
            account (Account): Account if successfully created; None if not
            errors (list): List of error messages in string form

        """

        account = None
        errors = []

        username = kwargs.get("username")
        password = kwargs.get("password")
        email = kwargs.get("email", "").strip()
        guest = kwargs.get("guest", False)

        permissions = kwargs.get("permissions", settings.PERMISSION_ACCOUNT_DEFAULT)
        typeclass = kwargs.get("typeclass", cls)

        ip = kwargs.get("ip", "")
        if isinstance(ip, (tuple, list)):
            ip = ip[0]

        if ip and CREATION_THROTTLE.check(ip):
            errors.append(
                _("You are creating too many accounts. Please log into an existing account.")
            )
            return None, errors

        # Normalize username
        username = cls.normalize_username(username)

        # Validate username
        if not guest:
            valid, errs = cls.validate_username(username)
            if not valid:
                # this echoes the restrictions made by django's auth
                # module (except not allowing spaces, for convenience of
                # logging in).
                errors.extend(errs)
                return None, errors

        # Validate password
        # Have to create a dummy Account object to check username similarity
        valid, errs = cls.validate_password(password, account=cls(username=username))
        if not valid:
            errors.extend(errs)
            return None, errors

        # Check IP and/or name bans
        banned = cls.is_banned(username=username, ip=ip)
        if banned:
            # this is a banned IP or name!
            string = _(
                "|rYou have been banned and cannot continue from here."
                "\nIf you feel this ban is in error, please email an admin.|x"
            )
            errors.append(string)
            return None, errors

        # everything's ok. Create the new account.
        try:
            try:
                account = create.create_account(
                    username, email, password, permissions=permissions, typeclass=typeclass
                )
                logger.log_sec(f"Account Created: {account} (IP: {ip}).")

            except Exception:
                errors.append(
                    _(
                        "There was an error creating the Account. "
                        "If this problem persists, contact an admin."
                    )
                )
                logger.log_trace()
                return None, errors

            # This needs to be set so the engine knows this account is
            # logging in for the first time. (so it knows to call the right
            # hooks during login later)
            account.db.FIRST_LOGIN = True

            # Record IP address of creation, if available
            if ip:
                account.db.creator_ip = ip

            # join the new account to the public channels
            for chan_info in settings.DEFAULT_CHANNELS:
                if chankey := chan_info.get("key"):
                    channel = ChannelDB.objects.get_channel(chankey)
                    if not channel or not (
                        channel.access(account, "listen") and channel.connect(account)
                    ):
                        string = (
                            f"New account '{account.key}' could not connect to default channel"
                            f" '{chankey}'!"
                        )
                        logger.log_err(string)
                else:
                    logger.log_err(f"Default channel '{chan_info}' is missing a 'key' field!")

            if account and _AUTO_CREATE_CHARACTER_WITH_ACCOUNT:
                # Auto-create a character to go with this account

                character, errs = account.create_character(
                    typeclass=kwargs.get("character_typeclass", account.default_character_typeclass)
                )
                if errs:
                    errors.extend(errs)

        except Exception:
            # We are in the middle between logged in and -not, so we have
            # to handle tracebacks ourselves at this point. If we don't,
            # we won't see any errors at all.
            errors.append(_("An error occurred. Please e-mail an admin if the problem persists."))
            logger.log_trace()

        # Update the throttle to indicate a new account was created from this IP
        if ip and not guest:
            CREATION_THROTTLE.update(ip, "Too many accounts being created.")
        SIGNAL_ACCOUNT_POST_CREATE.send(sender=account, ip=ip)
        return account, errors

    def delete(self, *args, **kwargs):
        """
        Deletes the account persistently.

        Notes:
            `*args` and `**kwargs` are passed on to the base delete
             mechanism (these are usually not used).

        Return:
            bool: If deletion was successful. Only time it fails would be
                if the Account was already deleted. Note that even on a failure,
                connected resources (nicks/aliases etc) will still have been
                deleted.

        """
        for session in self.sessions.all():
            # unpuppeting all objects and disconnecting the user, if any
            # sessions remain (should usually be handled from the
            # deleting command)
            try:
                self.unpuppet_object(session)
            except RuntimeError:
                # no puppet to disconnect from
                pass
            session.sessionhandler.disconnect(session, reason=_("Account being deleted."))
        self.scripts.stop()
        self.attributes.clear()
        self.nicks.clear()
        self.aliases.clear()
        if not self.pk:
            return False
        super().delete(*args, **kwargs)
        return True

    # methods inherited from database model

    def msg(self, text=None, from_obj=None, session=None, options=None, **kwargs):
        """
        Evennia -> User
        This is the main route for sending data back to the user from the
        server.

        Args:
            text (str or tuple, optional): The message to send. This
                is treated internally like any send-command, so its
                value can be a tuple if sending multiple arguments to
                the `text` oob command.
            from_obj (Object or Account or list, optional): Object sending. If given, its
                at_msg_send() hook will be called. If iterable, call on all entities.
            session (Session or list, optional): Session object or a list of
                Sessions to receive this send. If given, overrules the
                default send behavior for the current
                MULTISESSION_MODE.
            options (list): Protocol-specific options. Passed on to the protocol.
        Keyword Args:
            any (dict): All other keywords are passed on to the protocol.

        """
        if from_obj:
            # call hook
            for obj in make_iter(from_obj):
                try:
                    obj.at_msg_send(text=text, to_obj=self, **kwargs)
                except Exception:
                    # this may not be assigned.
                    logger.log_trace()
        try:
            if not self.at_msg_receive(text=text, **kwargs):
                # abort message to this account
                return
        except Exception:
            # this may not be assigned.
            pass

        kwargs["options"] = options

        if text is not None:
            if not (isinstance(text, str) or isinstance(text, tuple)):
                # sanitize text before sending across the wire
                try:
                    text = to_str(text)
                except Exception:
                    text = repr(text)
            kwargs["text"] = text

        # session relay
        sessions = make_iter(session) if session else self.sessions.all()
        for session in sessions:
            session.data_out(**kwargs)

    def execute_cmd(self, raw_string, session=None, **kwargs):
        """
        Do something as this account. This method is never called normally,
        but only when the account object itself is supposed to execute the
        command. It takes account nicks into account, but not nicks of
        eventual puppets.

        Args:
            raw_string (str): Raw command input coming from the command line.
            session (Session, optional): The session to be responsible
                for the command-send

        Keyword Args:
            kwargs (any): Other keyword arguments will be added to the
                found command object instance as variables before it
                executes. This is unused by default Evennia but may be
                used to set flags and change operating parameters for
                commands at run-time.

        """
        # break circular import issues
        global _CMDHANDLER
        if not _CMDHANDLER:
            from evennia.commands.cmdhandler import cmdhandler as _CMDHANDLER
        raw_string = self.nicks.nickreplace(
            raw_string, categories=("inputline", "channel"), include_account=False
        )
        if not session and _MULTISESSION_MODE in (0, 1):
            # for these modes we use the first/only session
            sessions = self.sessions.get()
            session = sessions[0] if sessions else None

        return _CMDHANDLER(self, raw_string, callertype="account", session=session, **kwargs)

    # channel receive hooks

    def at_pre_channel_msg(self, message, channel, senders=None, **kwargs):
        """
        Called by the Channel just before passing a message into `channel_msg`.
        This allows for tweak messages per-user and also to abort the
        receive on the receiver-level.

        Args:
            message (str): The message sent to the channel.
            channel (Channel): The sending channel.
            senders (list, optional): Accounts or Objects acting as senders.
                For most normal messages, there is only a single sender. If
                there are no senders, this may be a broadcasting message.
            **kwargs: These are additional keywords passed into `channel_msg`.
                If `no_prefix=True` or `emit=True` are passed, the channel
                prefix will not be added (`[channelname]: ` by default)

        Returns:
            str or None: Allows for customizing the message for this recipient.
                If returning `None` (or `False`) message-receiving is aborted.
                The returning string will be passed into `self.channel_msg`.

        Notes:
            This support posing/emotes by starting channel-send with : or ;.

        """
        if senders:
            sender_string = ", ".join(sender.get_display_name(self) for sender in senders)
            message_lstrip = message.lstrip()
            if message_lstrip.startswith((":", ";")):
                # this is a pose, should show as e.g. "User1 smiles to channel"
                spacing = "" if message_lstrip[1:].startswith((":", "'", ",")) else " "
                message = f"{sender_string}{spacing}{message_lstrip[1:]}"
            else:
                # normal message
                message = f"{sender_string}: {message}"

        if not kwargs.get("no_prefix") and not kwargs.get("emit"):
            message = channel.channel_prefix() + message

        return message

    def channel_msg(self, message, channel, senders=None, **kwargs):
        """
        This performs the actions of receiving a message to an un-muted
        channel.

        Args:
            message (str): The message sent to the channel.
            channel (Channel): The sending channel.
            senders (list, optional): Accounts or Objects acting as senders.
                For most normal messages, there is only a single sender. If
                there are no senders, this may be a broadcasting message or
                similar.
            **kwargs: These are additional keywords originally passed into
                `Channel.msg`.

        Notes:
            Before this, `Channel.at_pre_channel_msg` will fire, which offers a way
            to customize the message for the receiver on the channel-level.

        """
        self.msg(
            text=(message, {"from_channel": channel.id}),
            from_obj=senders,
            options={"from_channel": channel.id},
        )

    def at_post_channel_msg(self, message, channel, senders=None, **kwargs):
        """
        Called by `self.channel_msg` after message was received.

        Args:
            message (str): The message sent to the channel.
            channel (Channel): The sending channel.
            senders (list, optional): Accounts or Objects acting as senders.
                For most normal messages, there is only a single sender. If
                there are no senders, this may be a broadcasting message.
            **kwargs: These are additional keywords passed into `channel_msg`.

        """
        pass

    # search method

    def search(
        self,
        searchdata,
        return_puppet=False,
        search_object=False,
        typeclass=None,
        nofound_string=None,
        multimatch_string=None,
        use_nicks=True,
        quiet=False,
        **kwargs,
    ):
        """
        This is similar to `DefaultObject.search` but defaults to searching
        for Accounts only.

        Args:
            searchdata (str or int): Search criterion, the Account's
                key or dbref to search for.
            return_puppet (bool, optional): Instructs the method to
                return matches as the object the Account controls rather
                than the Account itself (or None) if nothing is puppeted).
            search_object (bool, optional): Search for Objects instead of
                Accounts. This is used by e.g. the @examine command when
                wanting to examine Objects while OOC.
            typeclass (Account typeclass, optional): Limit the search
                only to this particular typeclass. This can be used to
                limit to specific account typeclasses or to limit the search
                to a particular Object typeclass if `search_object` is True.
            nofound_string (str, optional): A one-time error message
                to echo if `searchdata` leads to no matches. If not given,
                will fall back to the default handler.
            multimatch_string (str, optional): A one-time error
                message to echo if `searchdata` leads to multiple matches.
                If not given, will fall back to the default handler.
            use_nicks (bool, optional): Use account-level nick replacement.
            quiet (bool, optional): If set, will not show any error to the user,
                and will also lead to returning a list of matches.

        Return:
            match (Account, Object or None): A single Account or Object match.
            list: If `quiet=True` this is a list of 0, 1 or more Account or Object matches.

        Notes:
            Extra keywords are ignored, but are allowed in call in
            order to make API more consistent with
            objects.objects.DefaultObject.search.

        """
        # handle me, self and *me, *self
        if isinstance(searchdata, str):
            # handle wrapping of common terms
            if searchdata.lower() in ("me", "*me", "self", "*self"):
                return self
        searchdata = self.nicks.nickreplace(
            searchdata, categories=("account",), include_account=False
        )
        if search_object:
            matches = ObjectDB.objects.object_search(searchdata, typeclass=typeclass)
        else:
            matches = AccountDB.objects.account_search(searchdata, typeclass=typeclass)

        if quiet:
            matches = list(matches)
            if return_puppet:
                matches = [match.puppet for match in matches]
        else:
            matches = _AT_SEARCH_RESULT(
                matches,
                self,
                query=searchdata,
                nofound_string=nofound_string,
                multimatch_string=multimatch_string,
            )
            if matches and return_puppet:
                try:
                    matches = matches.puppet
                except AttributeError:
                    return None
        return matches

    def access(
        self, accessing_obj, access_type="read", default=False, no_superuser_bypass=False, **kwargs
    ):
        """
        Determines if another object has permission to access this
        object in whatever way.

        Args:
          accessing_obj (Object): Object trying to access this one.
          access_type (str, optional): Type of access sought.
          default (bool, optional): What to return if no lock of
            access_type was found
          no_superuser_bypass (bool, optional): Turn off superuser
            lock bypassing. Be careful with this one.

        Keyword Args:
          kwargs (any): Passed to the at_access hook along with the result.

        Returns:
            result (bool): Result of access check.

        """
        result = super().access(
            accessing_obj,
            access_type=access_type,
            default=default,
            no_superuser_bypass=no_superuser_bypass,
        )
        self.at_access(result, accessing_obj, access_type, **kwargs)
        return result

    @property
    def idle_time(self):
        """
        Returns the idle time of the least idle session in seconds. If
        no sessions are connected it returns nothing.
        """
        idle = [session.cmd_last_visible for session in self.sessions.all()]
        if idle:
            return time.time() - float(max(idle))
        return None

    @property
    def connection_time(self):
        """
        Returns the maximum connection time of all connected sessions
        in seconds. Returns nothing if there are no sessions.
        """
        conn = [session.conn_time for session in self.sessions.all()]
        if conn:
            return time.time() - float(min(conn))
        return None

    # account hooks

    def basetype_setup(self):
        """
        This sets up the basic properties for an account. Overload this
        with at_account_creation rather than changing this method.

        """
        # A basic security setup
        self.locks.add(self.lockstring)

        # The ooc account cmdset
        self.cmdset.add_default(_CMDSET_ACCOUNT, persistent=True)

    def at_account_creation(self):
        """
        This is called once, the very first time the account is created
        (i.e. first time they register with the game). It's a good
        place to store attributes all accounts should have, like
        configuration values etc.

        """
        # set an (empty) attribute holding the characters this account has
        lockstring = "attrread:perm(Admins);attredit:perm(Admins);attrcreate:perm(Admins);"
        self.attributes.add("_playable_characters", [], lockstring=lockstring)
        self.attributes.add("_saved_protocol_flags", {}, lockstring=lockstring)

    def at_init(self):
        """
        This is always called whenever this object is initiated --
        that is, whenever it its typeclass is cached from memory. This
        happens on-demand first time the object is used or activated
        in some way after being created but also after each server
        restart or reload. In the case of account objects, this usually
        happens the moment the account logs in or reconnects after a
        reload.

        """
        pass

    # Note that the hooks below also exist in the character object's
    # typeclass. You can often ignore these and rely on the character
    # ones instead, unless you are implementing a multi-character game
    # and have some things that should be done regardless of which
    # character is currently connected to this account.

    def at_first_save(self):
        """
        This is a generic hook called by Evennia when this object is
        saved to the database the very first time.  You generally
        don't override this method but the hooks called by it.

        """
        self.basetype_setup()
        self.at_account_creation()
        # initialize Attribute/TagProperties
        self.init_evennia_properties()

        permissions = [settings.PERMISSION_ACCOUNT_DEFAULT]
        if hasattr(self, "_createdict"):
            # this will only be set if the utils.create_account
            # function was used to create the object.
            cdict = self._createdict
            updates = []
            if not cdict.get("key"):
                if not self.db_key:
                    self.db_key = f"#{self.dbid}"
                    updates.append("db_key")
            elif self.key != cdict.get("key"):
                updates.append("db_key")
                self.db_key = cdict["key"]
            if updates:
                self.save(update_fields=updates)

            if cdict.get("locks"):
                self.locks.add(cdict["locks"])
            if cdict.get("permissions"):
                permissions = cdict["permissions"]
            if cdict.get("tags"):
                # this should be a list of tags, tuples (key, category) or (key, category, data)
                self.tags.batch_add(*cdict["tags"])
            if cdict.get("attributes"):
                # this should be tuples (key, val, ...)
                self.attributes.batch_add(*cdict["attributes"])
            if cdict.get("nattributes"):
                # this should be a dict of nattrname:value
                for key, value in cdict["nattributes"]:
                    self.nattributes.add(key, value)
            del self._createdict

        self.permissions.batch_add(*permissions)

    def at_access(self, result, accessing_obj, access_type, **kwargs):
        """
        This is triggered after an access-call on this Account has
            completed.

        Args:
            result (bool): The result of the access check.
            accessing_obj (any): The object requesting the access
                check.
            access_type (str): The type of access checked.

        Keyword Args:
            kwargs (any): These are passed on from the access check
                and can be used to relay custom instructions from the
                check mechanism.

        Notes:
            This method cannot affect the result of the lock check and
            its return value is not used in any way. It can be used
            e.g.  to customize error messages in a central location or
            create other effects based on the access result.

        """
        pass

    def at_cmdset_get(self, **kwargs):
        """
        Called just before cmdsets on this object are requested by the
        command handler. If changes need to be done on the fly to the
        cmdset before passing them on to the cmdhandler, this is the
        place to do it. This is called also if the object currently
        have no cmdsets.

        Keyword Args:
            caller (Object, Account or Session): The object requesting the cmdsets.
            current (CmdSet): The current merged cmdset.
            force_init (bool): If `True`, force a re-build of the cmdset. (seems unused)
            **kwargs: Arbitrary input for overloads.

        """
        pass

    def get_cmdsets(self, caller, current, **kwargs):
        """
        Called by the CommandHandler to get a list of cmdsets to merge.

        Args:
            caller (obj): The object requesting the cmdsets.
            current (cmdset): The current merged cmdset.
            **kwargs: Arbitrary input for overloads.

        Returns:
            tuple: A tuple of (current, cmdsets), which is probably self.cmdset.current and self.cmdset.cmdset_stack
        """
        return self.cmdset.current, list(self.cmdset.cmdset_stack)

    def at_first_login(self, **kwargs):
        """
        Called the very first time this account logs into the game.
        Note that this is called *before* at_pre_login, so no session
        is established and usually no character is yet assigned at
        this point. This hook is intended for account-specific setup
        like configurations.

        Args:
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        pass

    def at_password_change(self, **kwargs):
        """
        Called after a successful password set/modify.

        Args:
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        pass

    def at_pre_login(self, **kwargs):
        """
        Called every time the user logs in, just before the actual
        login-state is set.

        Args:
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        pass

    def _send_to_connect_channel(self, message):
        """
        Helper method for loading and sending to the comm channel dedicated to
        connection messages. This will also be sent to the mudinfo channel.

        Args:
            message (str): A message to send to the connect channel.

        """
        global _MUDINFO_CHANNEL, _CONNECT_CHANNEL
        if _MUDINFO_CHANNEL is None:
            if settings.CHANNEL_MUDINFO:
                try:
                    _MUDINFO_CHANNEL = ChannelDB.objects.get(db_key=settings.CHANNEL_MUDINFO["key"])
                except ChannelDB.DoesNotExist:
                    logger.log_trace()
            else:
                _MUDINFO = False
        if _CONNECT_CHANNEL is None:
            if settings.CHANNEL_CONNECTINFO:
                try:
                    _CONNECT_CHANNEL = ChannelDB.objects.get(
                        db_key=settings.CHANNEL_CONNECTINFO["key"]
                    )
                except ChannelDB.DoesNotExist:
                    logger.log_trace()
            else:
                _CONNECT_CHANNEL = False

        if settings.USE_TZ:
            now = timezone.localtime()
        else:
            now = timezone.now()
        now = "%02i-%02i-%02i(%02i:%02i)" % (now.year, now.month, now.day, now.hour, now.minute)
        if _MUDINFO_CHANNEL:
            _MUDINFO_CHANNEL.msg(f"[{now}]: {message}")
        if _CONNECT_CHANNEL:
            _CONNECT_CHANNEL.msg(f"[{now}]: {message}")

    def at_post_login(self, session=None, **kwargs):
        """
        Called at the end of the login process, just before letting
        the account loose.

        Args:
            session (Session, optional): Session logging in, if any.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Notes:
            This is called *before* an eventual Character's
            `at_post_login` hook. By default it is used to set up
            auto-puppeting based on `MULTISESSION_MODE`

        """
        # if we have saved protocol flags on ourselves, load them here.
        protocol_flags = self.attributes.get("_saved_protocol_flags", {})
        if session and protocol_flags:
            session.update_flags(**protocol_flags)

        # inform the client that we logged in through an OOB message
        if session:
            session.msg(logged_in={})

        self._send_to_connect_channel(_("|G{key} connected|n").format(key=self.key))
        if _AUTO_PUPPET_ON_LOGIN:
            # in this mode we try to auto-connect to our last connected object, if any
            try:
                self.puppet_object(session, self.db._last_puppet)
            except RuntimeError:
                self.msg(_("The Character does not exist."))
                return
        else:
            # In this mode we don't auto-connect but by default end up at a character selection
            # screen. We execute look on the account.
            self.msg(self.at_look(target=self.characters, session=session), session=session)

    def at_failed_login(self, session, **kwargs):
        """
        Called by the login process if a user account is targeted correctly
        but provided with an invalid password. By default it does nothing,
        but exists to be overridden.

        Args:
            session (session): Session logging in.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).
        """
        pass

    def at_disconnect(self, reason=None, **kwargs):
        """
        Called just before user is disconnected.

        Args:
            reason (str, optional): The reason given for the disconnect,
                (echoed to the connection channel by default).
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).


        """
        reason = f" ({reason if reason else ''})"
        self._send_to_connect_channel(
            _("|R{key} disconnected{reason}|n").format(key=self.key, reason=reason)
        )

    def at_post_disconnect(self, **kwargs):
        """
        This is called *after* disconnection is complete. No messages
        can be relayed to the account from here. After this call, the
        account should not be accessed any more, making this a good
        spot for deleting it (in the case of a guest account account,
        for example).

        Args:
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        pass

    def at_msg_receive(self, text=None, from_obj=None, **kwargs):
        """
        This hook is called whenever someone sends a message to this
        object using the `msg` method.

        Note that from_obj may be None if the sender did not include
        itself as an argument to the obj.msg() call - so you have to
        check for this. .

        Consider this a pre-processing method before msg is passed on
        to the user session. If this method returns False, the msg
        will not be passed on.

        Args:
            text (str, optional): The message received.
            from_obj (any, optional): The object sending the message.

        Keyword Args:
            This includes any keywords sent to the `msg` method.

        Returns:
            receive (bool): If this message should be received.

        Notes:
            If this method returns False, the `msg` operation
            will abort without sending the message.

        """
        return True

    def at_msg_send(self, text=None, to_obj=None, **kwargs):
        """
        This is a hook that is called when *this* object sends a
        message to another object with `obj.msg(text, to_obj=obj)`.

        Args:
            text (str, optional): Text to send.
            to_obj (any, optional): The object to send to.

        Keyword Args:
            Keywords passed from msg()

        Notes:
            Since this method is executed by `from_obj`, if no `from_obj`
            was passed to `DefaultCharacter.msg` this hook will never
            get called.

        """
        pass

    def at_server_reload(self):
        """
        This hook is called whenever the server is shutting down for
        restart/reboot. If you want to, for example, save
        non-persistent properties across a restart, this is the place
        to do it.
        """
        pass

    def at_server_shutdown(self):
        """
        This hook is called whenever the server is shutting down fully
        (i.e. not for a restart).
        """
        pass

    ooc_appearance_template = """
--------------------------------------------------------------------
{header}

{sessions}

  |whelp|n - more commands
  |wpublic <text>|n - talk on public channel
  |wcharcreate <name> [=description]|n - create new character
  |wchardelete <name>|n - delete a character
  |wic <name>|n - enter the game as character (|wooc|n to get back here)
  |wic|n - enter the game as latest character controlled.

{characters}
{footer}
--------------------------------------------------------------------
""".strip()

    def at_look(self, target=None, session=None, **kwargs):
        """
        Called when this object executes a look. It allows to customize
        just what this means.

        Args:
            target (Object or list, optional): An object or a list
                objects to inspect. This is normally a list of characters.
            session (Session, optional): The session doing this look.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            look_string (str): A prepared look string, ready to send
                off to any recipient (usually to ourselves)

        """

        if target and not is_iter(target):
            # single target - just show it
            if hasattr(target, "return_appearance"):
                return target.return_appearance(self)
            else:
                return f"{target} has no in-game appearance."

        # multiple targets - this is a list of characters
        characters = list(tar for tar in target if tar) if target else []
        ncars = len(characters)
        sessions = self.sessions.all()
        nsess = len(sessions)

        if not nsess:
            # no sessions, nothing to report
            return ""

        # header text
        txt_header = f"Account |g{self.name}|n (you are Out-of-Character)"

        # sessions
        sess_strings = []
        for isess, sess in enumerate(sessions):
            ip_addr = sess.address[0] if isinstance(sess.address, tuple) else sess.address
            addr = f"{sess.protocol_key} ({ip_addr})"
            sess_str = (
                f"|w* {isess + 1}|n"
                if session and session.sessid == sess.sessid
                else f"  {isess + 1}"
            )

            sess_strings.append(f"{sess_str} {addr}")

        txt_sessions = "|wConnected session(s):|n\n" + "\n".join(sess_strings)

        if not characters:
            txt_characters = "You don't have a character yet. Use |wcharcreate|n."
        else:
            max_chars = (
                "unlimited"
                if self.is_superuser or _MAX_NR_CHARACTERS is None
                else _MAX_NR_CHARACTERS
            )

            char_strings = []
            for char in characters:
                csessions = char.sessions.all()
                if csessions:
                    for sess in csessions:
                        # character is already puppeted
                        sid = sess in sessions and sessions.index(sess) + 1
                        if sess and sid:
                            char_strings.append(
                                f" - |G{char.name}|n [{', '.join(char.permissions.all())}] "
                                f"(played by you in session {sid})"
                            )
                        else:
                            char_strings.append(
                                f" - |R{char.name}|n [{', '.join(char.permissions.all())}] "
                                "(played by someone else)"
                            )
                else:
                    # character is "free to puppet"
                    char_strings.append(f" - {char.name} [{', '.join(char.permissions.all())}]")

            txt_characters = (
                f"Available character(s) ({ncars}/{max_chars}, |wic <name>|n to play):|n\n"
                + "\n".join(char_strings)
            )
        return self.ooc_appearance_template.format(
            header=txt_header,
            sessions=txt_sessions,
            characters=txt_characters,
            footer="",
        )


class DefaultGuest(DefaultAccount):
    """
    This class is used for guest logins. Unlike Accounts, Guests and
    their characters are deleted after disconnection.

    """

    @classmethod
    def create(cls, **kwargs):
        """
        Forwards request to cls.authenticate(); returns a DefaultGuest object
        if one is available for use.

        """
        return cls.authenticate(**kwargs)

    @classmethod
    def authenticate(cls, **kwargs):
        """
        Gets or creates a Guest account object.

        Keyword Args:
            ip (str, optional): IP address of requester; used for ban checking,
                throttling and logging

        Returns:
            account (Object): Guest account object, if available
            errors (list): List of error messages accrued during this request.

        """
        errors = []
        account = None
        username = None
        ip = kwargs.get("ip", "").strip()

        # check if guests are enabled.
        if not settings.GUEST_ENABLED:
            errors.append(_("Guest accounts are not enabled on this server."))
            return None, errors

        try:
            # Find an available guest name.
            for name in settings.GUEST_LIST:
                if not AccountDB.objects.filter(username__iexact=name).exists():
                    username = name
                    break
            if not username:
                errors.append(_("All guest accounts are in use. Please try again later."))
                if ip:
                    LOGIN_THROTTLE.update(ip, "Too many requests for Guest access.")
                return None, errors
            else:
                # build a new account with the found guest username
                password = "%016x" % getrandbits(64)
                home = settings.GUEST_HOME
                permissions = settings.PERMISSION_GUEST_DEFAULT
                typeclass = settings.BASE_GUEST_TYPECLASS

                # Call parent class creator
                account, errs = super(DefaultGuest, cls).create(
                    guest=True,
                    username=username,
                    password=password,
                    permissions=permissions,
                    typeclass=typeclass,
                    home=home,
                    ip=ip,
                )
                errors.extend(errs)

                if not account.characters:
                    # this can happen for multisession_mode > 1. For guests we
                    # always auto-create a character, regardless of multi-session-mode.
                    character, errs = account.create_character()

                if errs:
                    errors.extend(errs)

                return account, errors

        except Exception:
            # We are in the middle between logged in and -not, so we have
            # to handle tracebacks ourselves at this point. If we don't,
            # we won't see any errors at all.
            errors.append(_("An error occurred. Please e-mail an admin if the problem persists."))
            logger.log_trace()
            return None, errors

        return account, errors

    def at_post_login(self, session=None, **kwargs):
        """
        By default, Guests only have one character regardless of which
        MAX_NR_CHARACTERS we use. They also always auto-puppet a matching
        character and don't get a choice.

        Args:
            session (Session, optional): Session connecting.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        self._send_to_connect_channel(_("|G{key} connected|n").format(key=self.key))
        self.puppet_object(session, self.db._last_puppet)

    def at_server_shutdown(self):
        """
        We repeat the functionality of `at_disconnect()` here just to
        be on the safe side.
        """
        super().at_server_shutdown()
        for character in self.characters:
            character.delete()

    def at_post_disconnect(self, **kwargs):
        """
        Once having disconnected, destroy the guest's characters and

        Args:
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        super().at_post_disconnect()
        for character in self.characters:
            character.delete()
        self.delete()



# -*- coding: utf-8 -*-


from django.conf import settings
from django.db import migrations, models

import evennia.utils.picklefield


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Attribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID", serialize=False, auto_created=True, primary_key=True
                    ),
                ),
                ("db_key", models.CharField(max_length=255, verbose_name="key", db_index=True)),
                (
                    "db_value",
                    evennia.utils.picklefield.PickledObjectField(
                        help_text="The data returned when the attribute is accessed. Must be written as a Python literal if editing through the admin interface. Attribute values which are not Python literals cannot be edited through the admin interface.",
                        null=True,
                        verbose_name="value",
                    ),
                ),
                (
                    "db_strvalue",
                    models.TextField(
                        help_text="String-specific storage for quick look-up",
                        null=True,
                        verbose_name="strvalue",
                        blank=True,
                    ),
                ),
                (
                    "db_category",
                    models.CharField(
                        max_length=128,
                        blank=True,
                        help_text="Optional categorization of attribute.",
                        null=True,
                        verbose_name="category",
                        db_index=True,
                    ),
                ),
                (
                    "db_lock_storage",
                    models.TextField(
                        help_text="Lockstrings for this object are stored here.",
                        verbose_name="locks",
                        blank=True,
                    ),
                ),
                (
                    "db_model",
                    models.CharField(
                        max_length=32,
                        blank=True,
                        help_text="Which model of object this attribute is attached to (A natural key like objects.dbobject). You should not change this value unless you know what you are doing.",
                        null=True,
                        verbose_name="model",
                        db_index=True,
                    ),
                ),
                (
                    "db_attrtype",
                    models.CharField(
                        max_length=16,
                        blank=True,
                        help_text="Subclass of Attribute (None or nick)",
                        null=True,
                        verbose_name="attrtype",
                        db_index=True,
                    ),
                ),
                (
                    "db_date_created",
                    models.DateTimeField(auto_now_add=True, verbose_name="date_created"),
                ),
            ],
            options={"verbose_name": "Evennia Attribute"},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID", serialize=False, auto_created=True, primary_key=True
                    ),
                ),
                (
                    "db_key",
                    models.CharField(
                        help_text="tag identifier",
                        max_length=255,
                        null=True,
                        verbose_name="key",
                        db_index=True,
                    ),
                ),
                (
                    "db_category",
                    models.CharField(
                        help_text="tag category",
                        max_length=64,
                        null=True,
                        verbose_name="category",
                        db_index=True,
                    ),
                ),
                (
                    "db_data",
                    models.TextField(
                        help_text="optional data field with extra information. This is not searched for.",
                        null=True,
                        verbose_name="data",
                        blank=True,
                    ),
                ),
                (
                    "db_model",
                    models.CharField(
                        help_text="database model to Tag",
                        max_length=32,
                        null=True,
                        verbose_name="model",
                        db_index=True,
                    ),
                ),
                (
                    "db_tagtype",
                    models.CharField(
                        help_text="overall type of Tag",
                        max_length=16,
                        null=True,
                        verbose_name="tagtype",
                        db_index=True,
                    ),
                ),
            ],
            options={"verbose_name": "Tag"},
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="tag", unique_together=set([("db_key", "db_category", "db_tagtype")])
        ),
        migrations.AlterIndexTogether(
            name="tag", index_together=set([("db_key", "db_category", "db_tagtype")])
        ),
    ]
    # if we are using Oracle, we need to remove the AlterIndexTogether operation
    # since Oracle seems to create its own index already at AlterUniqueTogether, meaning
    # there is a conflict (see issue #732).
    if settings.DATABASES["default"]["ENGINE"] == "django.db.backends.oracle":
        del operations[3]



"""
Random string tests.

"""

from evennia.utils.test_resources import BaseEvenniaTest

from . import random_string_generator

SIMPLE_GENERATOR = random_string_generator.RandomStringGenerator("simple", "[01]{2}")


class TestRandomStringGenerator(BaseEvenniaTest):
    def test_generate(self):
        """Generate and fail when exhausted."""
        generated = []
        for i in range(4):
            generated.append(SIMPLE_GENERATOR.get())

        generated.sort()
        self.assertEqual(generated, ["00", "01", "10", "11"])

        # At this point, we have generated 4 strings.
        # We can't generate one more
        with self.assertRaises(random_string_generator.ExhaustedGenerator):
            SIMPLE_GENERATOR.get()



"""
Tags are entities that are attached to objects in the same way as
Attributes. But contrary to Attributes, which are unique to an
individual object, a single Tag can be attached to any number of
objects at the same time.

Tags are used for tagging, obviously, but the data structure is also
used for storing Aliases and Permissions. This module contains the
respective handlers.

"""
from collections import defaultdict

from django.conf import settings
from django.db import models

from evennia.locks.lockfuncs import perm as perm_lockfunc
from evennia.utils.utils import make_iter, to_str

_TYPECLASS_AGGRESSIVE_CACHE = settings.TYPECLASS_AGGRESSIVE_CACHE

# ------------------------------------------------------------
#
# Tags
#
# ------------------------------------------------------------


class Tag(models.Model):
    """
    Tags are quick markers for objects in-game. An typeobject can have
    any number of tags, stored via its db_tags property.  Tagging
    similar objects will make it easier to quickly locate the group
    later (such as when implementing zones). The main advantage of
    tagging as opposed to using tags is speed; a tag is very
    limited in what data it can hold, and the tag key+category is
    indexed for efficient lookup in the database. Tags are shared
    between objects - a new tag is only created if the key+category
    combination did not previously exist, making them unsuitable for
    storing object-related data (for this a regular Attribute should be
    used).

    The 'db_data' field is intended as a documentation field for the
    tag itself, such as to document what this tag+category stands for
    and display that in a web interface or similar.

    The main default use for Tags is to implement Aliases for objects.
    this uses the 'aliases' tag category, which is also checked by the
    default search functions of Evennia to allow quick searches by alias.

    """

    db_key = models.CharField(
        "key", max_length=255, null=True, help_text="tag identifier", db_index=True
    )
    db_category = models.CharField(
        "category", max_length=64, null=True, blank=True, help_text="tag category", db_index=True
    )
    db_data = models.TextField(
        "data",
        null=True,
        blank=True,
        help_text="optional data field with extra information. This is not searched for.",
    )
    # this is "objectdb" etc. Required behind the scenes
    db_model = models.CharField(
        "model", max_length=32, null=True, help_text="database model to Tag", db_index=True
    )
    # this is None, alias or permission
    db_tagtype = models.CharField(
        "tagtype",
        max_length=16,
        null=True,
        blank=True,
        help_text="overall type of Tag",
        db_index=True,
    )

    class Meta:
        "Define Django meta options"
        verbose_name = "Tag"
        unique_together = (("db_key", "db_category", "db_tagtype", "db_model"),)
        index_together = (("db_key", "db_category", "db_tagtype", "db_model"),)

    def __lt__(self, other):
        return str(self) < str(other)

    def __str__(self):
        return str(
            "<Tag: %s%s>"
            % (self.db_key, "(category:%s)" % self.db_category if self.db_category else "")
        )


#
# Handlers making use of the Tags model
#


class TagProperty:
    """
    Tag Property.
    """

    taghandler_name = "tags"

    def __init__(self, category=None, data=None):
        """
        Tag property descriptor. Allows for setting tags on an object as Django-like 'fields'
        on the class level. Since Tags are almost always used for querying, Tags are always
        created/assigned along with the object. Make sure the property/tagname does not collide
        with an existing method/property on the class. If it does, you must use tags.add()
        instead.

        Note that while you _can_ check e.g. `obj.tagname,this will give an AttributeError
        if the Tag is not set. Most often you want to use `obj.tags.get("tagname")` to check
        if a tag is set on an object.

        Example:
        ::

                class Character(DefaultCharacter):
                    mytag = TagProperty()  # category=None
                    mytag2 = TagProperty(category="tagcategory")
        """

        self._category = category
        self._data = data
        self._key = ""

    def __set_name__(self, cls, name):
        """
        Called when descriptor is first assigned to the class (not the instance!).
        It is called with the name of the field.

        """
        self._key = name

    def __get__(self, instance, owner):
        """
        Called when accessing the tag as a property on the instance.

        """
        try:
            return getattr(instance, self.taghandler_name).get(
                key=self._key, category=self._category, return_list=False, raise_exception=True
            )
        except AttributeError:
            self.__set__(instance, self._category)

    def __set__(self, instance, category):
        """
        Assign a new category to the tag. It's not possible to set 'data' this way.

        """
        self._category = category
        (
            getattr(instance, self.taghandler_name).add(
                key=self._key, category=self._category, data=self._data
            )
        )

    def __delete__(self, instance):
        """
        Called when running `del` on the property. Will disconnect the object from
        the Tag. Note that the tag will be readded on next fetch unless the
        TagProperty is also removed in code!

        """
        getattr(instance, self.taghandler_name).remove(key=self._key, category=self._category)


class TagCategoryProperty:
    """
    Tag Category Property.

    """

    taghandler_name = "tags"

    def __init__(self, *default_tags):
        """
        Assign a property for a Tag Category, with any number of Tag keys.
        This is often more useful than the `TagProperty` since it's common to want to check which
        tags of a particular category the object is a member of.

        Args:
            *args (str or callable): Tag keys to assign to this property, using the category given
                by the name of the property. Note that, if these tags are always set on the object,
                if they are removed by some other means, they will be re-added when this property
                is accessed. Furthermore, changing this list after the object was created, will
                not remove any old tags (there is no way for the property to know if the
                new list is new or not).  If a callable, it will be called without arguments to
                return the tag key. It is not possible to set tag `data` this way (use the
                Taghandler directly for that). Tag keys are not case sensitive.

        Raises:
            ValueError: If the input is not a valid tag key or tuple.

        Notes:
            It is not possible to set Tags with a `None` category using a `TagCategoryProperty` -
            use `obj.tags.add()` instead.

        Example:
        ::

                class RogueCharacter(DefaultCharacter):
                    guild = TagCategoryProperty("thieves_guild", "merchant_guild")

        """
        self._category = ""
        self._default_tags = self._parse_tag_input(*default_tags)

    def _parse_tag_input(self, *args):
        """
        Parse input to the property.

        Args:
            *args (str or callable): Tags, either as strings or `callable`, which should return
            the tag key when called without arguments. Keys are not case sensitive.

        Returns:
            list: A list of tag keys.

        """
        tags = []
        for tagkey in args:
            if callable(tagkey):
                tagkey = tagkey()
            tags.append((str(tagkey).lower()))
        return tags

    def __set_name__(self, cls, name):
        """
        Called when descriptor is first assigned to the class (not the instance!).
        It is called with the name of the field.

        """
        self._category = name

    def __get__(self, instance, owner):
        """
        Called when accessing the tag as a property on the instance. Returns a list
        of tags under the given category.
        """
        taghandler = getattr(instance, self.taghandler_name)

        default_tags = self._default_tags
        tags = taghandler.get(category=self._category, return_list=True)

        missing_default_tags = set(default_tags) - set(tags)

        if missing_default_tags:
            getattr(instance, self.taghandler_name).batch_add(
                *[(tag, self._category) for tag in missing_default_tags]
            )

            tags += missing_default_tags  # to avoid a second db call

        return tags

    def __set__(self, instance, *args):
        """
        Assign a new set of tags to the category. Note that we can't know if previous
        tags were assigned from this property or from TagHandler, so we don't
        remove old tags. To refresh to only have the tags in this constructor, first
        use `del` on this property and re-access the property with the changed default list.

        """
        getattr(instance, self.taghandler_name).batch_add(*[(tag, self._category) for tag in args])

    def __delete__(self, instance):
        """
        Called when running `del` on the property. Will remove all tags of this
        category from the object. Note that next time this desriptor is accessed, the
        default ones will be re-added!

        Note:
            This will remove _all_ tags of this category from the object. This is necessary
            in order to be able to be able to combine this with `__set__` to get a tag
            list where property and handler are in sync.

        """
        getattr(instance, self.taghandler_name).remove(category=self._category)


class TagHandler(object):
    """
    Generic tag-handler. Accessed via TypedObject.tags.

    """

    _m2m_fieldname = "db_tags"
    _tagtype = None

    def __init__(self, obj):
        """
        Tags are stored internally in the TypedObject.db_tags m2m
        field with an tag.db_model based on the obj the taghandler is
        stored on and with a tagtype given by self.handlertype

        Args:
            obj (object): The object on which the handler is set.

        """
        self.obj = obj
        self._objid = obj.id
        self._model = obj.__dbclass__.__name__.lower()
        self._cache = {}
        # store category names fully cached
        self._catcache = {}
        # full cache was run on all tags
        self._cache_complete = False

    def _query_all(self):
        """
        Get all tags for this object.

        """
        query = {
            "%s__id" % self._model: self._objid,
            "tag__db_model": self._model,
            "tag__db_tagtype": self._tagtype,
        }
        return [
            conn.tag
            for conn in getattr(self.obj, self._m2m_fieldname).through.objects.filter(**query)
        ]

    def _fullcache(self):
        """
        Cache all tags of this object.

        """
        if not _TYPECLASS_AGGRESSIVE_CACHE:
            return
        tags = self._query_all()
        self._cache = dict(
            (
                "%s-%s"
                % (
                    to_str(tag.db_key).lower(),
                    tag.db_category.lower() if tag.db_category else None,
                ),
                tag,
            )
            for tag in tags
        )
        self._cache_complete = True

    def _getcache(self, key=None, category=None):
        """
        Retrieve from cache or database (always caches)

        Args:
            key (str, optional): Tag key to query for
            category (str, optional): Tag category

        Returns:
            args (list): Returns a list of zero or more matches
                found from cache or database.
        Notes:
            When given a category only, a search for all objects
            of that category is done and a the category *name* is is
            stored. This tells the system on subsequent calls that the
            list of cached tags of this category is up-to-date
            and that the cache can be queried for category matches
            without missing any.
            The TYPECLASS_AGGRESSIVE_CACHE=False setting will turn off
            caching, causing each tag access to trigger a
            database lookup.

        """
        key = str(key).strip().lower() if key else None
        category = category.strip().lower() if category else None
        if key:
            cachekey = "%s-%s" % (key, category)
            tag = _TYPECLASS_AGGRESSIVE_CACHE and self._cache.get(cachekey, None)
            if tag and (not hasattr(tag, "pk") and tag.pk is None):
                # clear out Tags deleted from elsewhere. We must search this anew.
                tag = None
                del self._cache[cachekey]
            if tag:
                return [tag]  # return cached entity
            else:
                query = {
                    "%s__id" % self._model: self._objid,
                    "tag__db_model": self._model,
                    "tag__db_tagtype": self._tagtype,
                    "tag__db_key__iexact": key.lower(),
                    "tag__db_category__iexact": category.lower() if category else None,
                }
                conn = getattr(self.obj, self._m2m_fieldname).through.objects.filter(**query)
                if conn:
                    tag = conn[0].tag
                    if _TYPECLASS_AGGRESSIVE_CACHE:
                        self._cache[cachekey] = tag
                    return [tag]
        else:
            # only category given (even if it's None) - we can't
            # assume the cache to be complete unless we have queried
            # for this category before
            catkey = "-%s" % category
            if _TYPECLASS_AGGRESSIVE_CACHE and catkey in self._catcache:
                return [tag for key, tag in self._cache.items() if key.endswith(catkey)]
            else:
                # we have to query to make this category up-date in the cache
                query = {
                    "%s__id" % self._model: self._objid,
                    "tag__db_model": self._model,
                    "tag__db_tagtype": self._tagtype,
                    "tag__db_category__iexact": category.lower() if category else None,
                }
                tags = [
                    conn.tag
                    for conn in getattr(self.obj, self._m2m_fieldname).through.objects.filter(
                        **query
                    )
                ]
                if _TYPECLASS_AGGRESSIVE_CACHE:
                    for tag in tags:
                        cachekey = "%s-%s" % (tag.db_key, category)
                        self._cache[cachekey] = tag
                    # mark category cache as up-to-date
                    self._catcache[catkey] = True
                return tags
        return []

    def _setcache(self, key, category, tag_obj):
        """
        Update cache.

        Args:
            key (str): A cleaned key string
            category (str or None): A cleaned category name
            tag_obj (tag): The newly saved tag

        """
        if not _TYPECLASS_AGGRESSIVE_CACHE:
            return
        if not key:  # don't allow an empty key in cache
            return
        key, category = (
            str(key).strip().lower(),
            category.strip().lower() if category else category,
        )
        cachekey = "%s-%s" % (key, category)
        catkey = "-%s" % category
        self._cache[cachekey] = tag_obj
        # mark that the category cache is no longer up-to-date
        self._catcache.pop(catkey, None)
        self._cache_complete = False

    def _delcache(self, key, category):
        """
        Remove tag from cache

        Args:
            key (str): A cleaned key string
            category (str or None): A cleaned category name

        """
        key, category = (
            str(key).strip().lower(),
            category.strip().lower() if category else category,
        )
        catkey = "-%s" % category
        if key:
            cachekey = "%s-%s" % (key, category)
            self._cache.pop(cachekey, None)
        else:
            [self._cache.pop(key, None) for key in self._cache if key.endswith(catkey)]
        # mark that the category cache is no longer up-to-date
        self._catcache.pop(catkey, None)
        self._cache_complete = False

    def reset_cache(self):
        """
        Reset the cache from the outside.

        """
        self._cache_complete = False
        self._cache = {}
        self._catcache = {}

    def add(self, key=None, category=None, data=None):
        """
        Add a new tag to the handler.

        Args:
            key (str or list): The name of the tag to add. If a list,
                add several Tags.
            category (str, optional): Category of Tag. `None` is the default category.
            data (str, optional): Info text about the tag(s) added.
                This can not be used to store object-unique info but only
                eventual info about the tag itself.

        Notes:
            If the tag + category combination matches an already
            existing Tag object, this will be re-used and no new Tag
            will be created.

        """
        if not key:
            return
        if not self._cache_complete:
            self._fullcache()
        for tagstr in make_iter(key):
            if not tagstr:
                continue
            tagstr = str(tagstr).strip().lower()
            category = str(category).strip().lower() if category else category
            data = str(data) if data is not None else None
            # this will only create tag if no matches existed beforehand (it
            # will overload data on an existing tag since that is not
            # considered part of making the tag unique)
            tagobj = self.obj.__class__.objects.create_tag(
                key=tagstr, category=category, data=data, tagtype=self._tagtype
            )
            getattr(self.obj, self._m2m_fieldname).add(tagobj)
            self._setcache(tagstr, category, tagobj)

    def has(self, key=None, category=None, return_list=False):
        """
        Checks if the given Tag (or list of Tags) exists on the object.

        Args:
            key (str or iterable): The Tag key or tags to check for.
                If `None`, search by category.
            category (str, optional): Limit the check to Tags with this
                category (note, that `None` is the default category).

        Returns:
            has_tag (bool or list): If the Tag exists on this object or not.
             If `tag` was given as an iterable then the return is a list of booleans.

        Raises:
            ValueError: If neither `tag` nor `category` is given.

        """
        ret = []
        category = category.strip().lower() if category is not None else None
        if key:
            for tag_str in make_iter(key):
                tag_str = str(tag_str).strip().lower()
                ret.append(bool(self._getcache(tag_str, category)))
        elif category:
            ret.extend(bool(tag) for tag in self._getcache(category=category))
        else:
            raise ValueError("Either tag or category must be provided.")

        if return_list:
            return ret

        return ret[0] if len(ret) == 1 else ret

    def get(
        self,
        key=None,
        default=None,
        category=None,
        return_tagobj=False,
        return_list=False,
        raise_exception=False,
    ):
        """
        Get the tag for the given key, category or combination of the two.

        Args:
            key (str or list, optional): The tag or tags to retrieve.
            default (any, optional): The value to return in case of no match.
            category (str, optional): The Tag category to limit the
                request to. Note that `None` is the valid, default
                category. If no `key` is given, all tags of this category will be
                returned.
            return_tagobj (bool, optional): Return the Tag object itself
                instead of a string representation of the Tag.
            return_list (bool, optional): Always return a list, regardless
                of number of matches.
            raise_exception (bool, optional): Raise AttributeError if no matches
                are found.

        Returns:
            tags (list): The matches, either string
                representations of the tags or the Tag objects themselves
                depending on `return_tagobj`. If 'default' is set, this
                will be a list with the default value as its only element.

        Raises:
            AttributeError: If finding no matches and `raise_exception` is True.

        """
        ret = []
        for keystr in make_iter(key):
            # note - the _getcache call removes case sensitivity for us
            ret.extend(
                [
                    tag if return_tagobj else to_str(tag.db_key)
                    for tag in self._getcache(keystr, category)
                ]
            )
        if not ret:
            if raise_exception:
                raise AttributeError(f"No tags found matching input {key}, {category}.")
            elif return_list:
                return [default] if default is not None else []
            else:
                return default
        return ret if return_list else (ret[0] if len(ret) == 1 else ret)

    def remove(self, key=None, category=None):
        """
        Remove a tag from the handler based ond key and/or category.

        Args:
            key (str or list, optional): The tag or tags to retrieve.
            category (str, optional): The Tag category to limit the
                request to. Note that `None` is the valid, default
                category
        Notes:
            If neither key nor category is specified, this acts
            as .clear().

        """
        if not key:
            # only category
            self.clear(category=category)
            return

        for key in make_iter(key):
            if not (key or key.strip()):  # we don't allow empty tags
                continue
            tagstr = str(key).strip().lower()
            category = category.strip().lower() if category else category

            # This does not delete the tag object itself. Maybe it should do
            # that when no objects reference the tag anymore (but how to check)?
            # For now, tags are never deleted, only their connection to objects.
            tagobj = getattr(self.obj, self._m2m_fieldname).filter(
                db_key=tagstr, db_category=category, db_model=self._model, db_tagtype=self._tagtype
            )
            if tagobj:
                getattr(self.obj, self._m2m_fieldname).remove(tagobj[0])
            self._delcache(key, category)

    def clear(self, category=None):
        """
        Remove all tags from the handler.

        Args:
            category (str, optional): The Tag category to limit the
                request to. Note that `None` is the valid, default
                category.

        """
        if not self._cache_complete:
            self._fullcache()
        query = {
            "%s__id" % self._model: self._objid,
            "tag__db_model": self._model,
            "tag__db_tagtype": self._tagtype,
        }
        if category:
            query["tag__db_category"] = category.strip().lower()
        getattr(self.obj, self._m2m_fieldname).through.objects.filter(**query).delete()
        self._cache = {}
        self._catcache = {}
        self._cache_complete = False

    def all(self, return_key_and_category=False, return_objs=False):
        """
        Get all tags in this handler, regardless of category.

        Args:
            return_key_and_category (bool, optional): Return a list of
                tuples `[(key, category), ...]`.
            return_objs (bool, optional): Return tag objects.

        Returns:
            tags (list): A list of tag keys `[tagkey, tagkey, ...]` or
                a list of tuples `[(key, category), ...]` if
                `return_key_and_category` is set.

        """
        if _TYPECLASS_AGGRESSIVE_CACHE:
            if not self._cache_complete:
                self._fullcache()
            tags = sorted(self._cache.values())
        else:
            tags = sorted(self._query_all())

        if return_key_and_category:
            # return tuple (key, category)
            return [(to_str(tag.db_key), tag.db_category) for tag in tags]
        elif return_objs:
            return tags
        else:
            return [to_str(tag.db_key) for tag in tags]

    def batch_add(self, *args):
        """
        Batch-add tags from a list of tuples.

        Args:
            *args (tuple or str): Each argument should be a `tagstr` keys or tuple
                `(keystr, category)` or `(keystr, category, data)`. It's possible to mix input
                types.

        Notes:
            This will generate a mimimal number of self.add calls,
            based on the number of categories involved (including
            `None`) (data is not unique and may be overwritten by the content
            of a latter tuple with the same category).

        """
        keys = defaultdict(list)
        data = {}
        for tup in args:
            tup = make_iter(tup)
            nlen = len(tup)
            if nlen == 1:  # just a key, no category
                keys[None].append(tup[0])
            elif nlen == 2:
                keys[tup[1]].append(tup[0])
            else:
                keys[tup[1]].append(tup[0])
                data[tup[1]] = tup[2]  # overwrite previous
        for category, key in keys.items():
            self.add(key=key, category=category, data=data.get(category, None))

    def batch_remove(self, *args):
        """
        Batch-remove tags from a list of tuples.

        Args:
            *args (tuple or str): Each argument should be a `tagstr` keys or tuple
                `(keystr, category)` or `(keystr, category, data)` (the `data` field is ignored,
                only `keystr`/`category` matters). It's possible to mix input types.

        """
        keys = defaultdict(list)
        for tup in args:
            tup = make_iter(tup)
            nlen = len(tup)
            if nlen == 1:  # just a key, no category
                keys[None].append(tup[0])
            elif nlen > 1:
                keys[tup[1]].append(tup[0])
        for category, key in keys.items():
            self.remove(key=key, category=category, data=data.get(category, None))

    def __str__(self):
        return ",".join(self.all())


class AliasProperty(TagProperty):
    """
    Allows for setting aliases like Django fields:
    ::

        class Character(DefaultCharacter):
            # note that every character will get the alias bob. Make sure
            # the alias property does not collide with an existing method
            # or property on the class.
            bob = AliasProperty()

    """

    taghandler_name = "aliases"


class AliasHandler(TagHandler):
    """
    A handler for the Alias Tag type.

    """

    _tagtype = "alias"


class PermissionProperty(TagProperty):
    """
    Allows for setting permissions like Django fields:
    ::

        class Character(DefaultCharacter):
            # note that every character will get this permission! Make
            # sure it doesn't collide with an existing method or property.
            myperm = PermissionProperty()

    """

    taghandler_name = "permissions"


class PermissionHandler(TagHandler):
    """
    A handler for the Permission Tag type.

    """

    _tagtype = "permission"

    def check(self, *permissions, require_all=False):
        """
        Straight-up check the provided permission against this handler. The check will pass if

        - any/all given permission exists on the handler (depending on if `require_all` is set).
        - If handler sits on puppeted object and this is a hierarachical perm, the puppeting
          Account's permission will also be included in the check, prioritizing the Account's perm
          (this avoids escalation exploits by puppeting a too-high prio character)
        - a permission is also considered to exist on the handler, if it is *lower* than
          a permission on the handler and this is a 'hierarchical' permission given
          in `settings.PERMISSION_HIERARCHY`. Example: If the 'Developer' hierarchical
          perm perm is set on the handler, and we check for the 'Builder' perm, the
          check will pass.

        Args:
            *permissions (str): Any number of permissions to check. By default,
                the permission is passed if any of these (or higher, if a
                hierarchical permission defined in settings.PERMISSION_HIERARCHY)
                exists in the handler. Permissions are not case-sensitive.
            require_all (bool): If set, *all* provided permissions much pass
                the check for the entire check to pass. By default only one
                needs to pass.

        Returns:
            bool: If the provided permission(s) pass the check on this handler.

        Example:
            ::
                can_enter = obj.permissions.check("Blacksmith", "Builder")

        Notes:
            This works the same way as the `perms` lockfunc and could be
            replicated with a lock check against the lockstring

                "locktype: perm(perm1) OR perm(perm2) OR ..."

            (using AND for the `require_all` condition).

        """
        if require_all:
            return all(perm_lockfunc(self.obj, None, perm) for perm in permissions)
        else:
            return any(perm_lockfunc(self.obj, None, perm) for perm in permissions)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2016-12-06 19:12


from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("comms", "0009_auto_20160921_1731")]

    operations = [
        migrations.AlterField(
            model_name="channeldb",
            name="db_object_subscriptions",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                null=True,
                related_name="object_subscription_set",
                to="objects.ObjectDB",
                verbose_name="subscriptions",
            ),
        ),
        migrations.AlterField(
            model_name="channeldb",
            name="db_subscriptions",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                null=True,
                related_name="subscription_set",
                to=settings.AUTH_USER_MODEL,
                verbose_name="subscriptions",
            ),
        ),
    ]



"""
Test NPC classes.

"""

from evennia import create_object
from evennia.utils.test_resources import EvenniaTest

from .. import npcs


class TestNPCBase(EvenniaTest):
    def test_npc_base(self):
        npc = create_object(
            npcs.EvAdventureNPC,
            key="TestNPC",
            attributes=[("hit_dice", 4), ("armor", 1), ("morale", 9)],
        )

        self.assertEqual(npc.hp_multiplier, 4)
        self.assertEqual(npc.hp_max, 16)
        self.assertEqual(npc.strength, 4)
        self.assertEqual(npc.charisma, 4)



import datetime

import git
from django.conf import settings

import evennia
from evennia import CmdSet, InterruptCommand
from evennia.commands.default.muxcommand import MuxCommand
from evennia.utils.utils import list_to_string


class GitCommand(MuxCommand):
    """
    The shared functionality between git/git evennia
    """

    def parse(self):
        """
        Parse the arguments, set default arg to 'status' and check for existence of currently targeted repo
        """
        super().parse()

        if self.args:
            split_args = self.args.strip().split(" ", 1)
            self.action = split_args[0]
            if len(split_args) > 1:
                self.args = "".join(split_args[1:])
            else:
                self.args = ""
        else:
            self.action = "status"
            self.args = ""

        self.err_msgs = [
            "|rInvalid Git Repository|n:",
            "The {repo_type} repository is not recognized as a git directory.",
            "In order to initialize it as a git directory, you will need to access your terminal and run the following commands from within your directory:",
            "    git init",
            "    git remote add origin {remote_link}",
        ]

        try:
            self.repo = git.Repo(self.directory, search_parent_directories=True)
        except git.exc.InvalidGitRepositoryError:
            err_msg = "\n".join(self.err_msgs).format(
                repo_type=self.repo_type, remote_link=self.remote_link
            )
            self.caller.msg(err_msg)
            raise InterruptCommand

        self.commit = self.repo.head.commit

        try:
            self.branch = self.repo.active_branch.name
        except TypeError as type_err:
            self.caller.msg(type_err)
            raise InterruptCommand

    def short_sha(self, repo, hexsha):
        """
        Utility: Get the short SHA of a commit.
        """
        short_sha = repo.git.rev_parse(hexsha, short=True)
        return short_sha

    def get_status(self):
        """
        Retrieves the status of the active git repository, displaying unstaged changes/untracked files.
        """
        time_of_commit = datetime.datetime.fromtimestamp(self.commit.committed_date)
        status_msg = "\n".join(
            [
                f"Branch: |w{self.branch}|n ({self.repo.git.rev_parse(self.commit.hexsha, short=True)}) ({time_of_commit})",
                f"By {self.commit.author.email}: {self.commit.message}",
            ]
        )

        changedFiles = {item.a_path for item in self.repo.index.diff(None)}
        if changedFiles:
            status_msg += f"Unstaged/uncommitted changes:|/    |g{'|/    '.join(changedFiles)}|n|/"
        if len(self.repo.untracked_files) > 0:
            status_msg += f"Untracked files:|/    |x{'|/    '.join(self.repo.untracked_files)}|n"
        return status_msg

    def get_branches(self):
        """
        Display current and available branches.
        """
        remote_refs = self.repo.remote().refs
        branch_msg = (
            f"Current branch: |w{self.branch}|n. Branches available: {list_to_string(remote_refs)}"
        )
        return branch_msg

    def checkout(self):
        """
        Check out a specific branch.
        """
        remote_refs = self.repo.remote().refs
        to_branch = self.args.strip().removeprefix(
            "origin/"
        )  # Slightly hacky, but git tacks on the origin/

        if to_branch not in remote_refs:
            self.caller.msg(f"Branch '{to_branch}' not available.")
            return False
        elif to_branch == self.branch:
            self.caller.msg(f"Already on |w{to_branch}|n. Maybe you want <git pull>?")
            return False
        else:
            try:
                self.repo.git.checkout(to_branch)
            except git.exc.GitCommandError as err:
                self.caller.msg("Couldn't checkout {} ({})".format(to_branch, err.stderr.strip()))
                return False
            self.msg(f"Checked out |w{to_branch}|n successfully. Server restart initiated.")
            return True

    def pull(self):
        """
        Attempt to pull new code.
        """
        old_commit = self.commit
        try:
            self.repo.remotes.origin.pull()
        except git.exc.GitCommandError as err:
            self.caller.msg("Couldn't pull {} ({})".format(self.branch, err.stderr.strip()))
            return False
        if old_commit == self.repo.head.commit:
            self.caller.msg("No new code to pull, no need to reset.\n")
            return False
        else:
            self.caller.msg(
                f"You have pulled new code. Server restart initiated.|/Head now at {self.repo.git.rev_parse(self.repo.head.commit.hexsha, short=True)}.|/Author: {self.repo.head.commit.author.name} ({self.repo.head.commit.author.email})|/{self.repo.head.commit.message.strip()}"
            )
            return True

    def func(self):
        """
        Provide basic Git functionality within the game.
        """
        caller = self.caller

        if self.action == "status":
            caller.msg(self.get_status())
        elif self.action == "branch" or (self.action == "checkout" and not self.args):
            caller.msg(self.get_branches())
        elif self.action == "checkout":
            if self.checkout():
                evennia.SESSION_HANDLER.portal_restart_server()
        elif self.action == "pull":
            if self.pull():
                evennia.SESSION_HANDLER.portal_restart_server()
        else:
            caller.msg("You can only git status, git branch, git checkout, or git pull.")
            return


class CmdGitEvennia(GitCommand):
    """
    Pull the latest code from the evennia core or checkout a different branch.

    Usage:
        git evennia status - View an overview of the evennia repository status.
        git evennia branch - View available branches in evennia.
        git evennia checkout <branch> - Checkout a different branch in evennia.
        git evennia pull - Pull the latest evennia code.

    For updating your local mygame repository, the same commands are available with 'git'.

    If there are any conflicts encountered, the command will abort. The command will reload your game after pulling new code automatically, but for some changes involving persistent scripts etc, you may need to manually restart.
    """

    key = "git evennia"
    locks = "cmd:pperm(Developer)"
    help_category = "System"
    directory = settings.EVENNIA_DIR
    repo_type = "Evennia"
    remote_link = "https://github.com/evennia/evennia.git"


class CmdGit(GitCommand):
    """
    Pull the latest code from your repository or checkout a different branch.

    Usage:
        git status        - View an overview of your git repository.
        git branch        - View available branches.
        git checkout main - Checkout the main branch of your code.
        git pull          - Pull the latest code from your current branch.

    For updating evennia code, the same commands are available with 'git evennia'.

    If there are any conflicts encountered, the command will abort. The command will reload your game after pulling new code automatically, but for changes involving persistent scripts etc, you may need to manually restart.
    """

    key = "git"
    locks = "cmd:pperm(Developer)"
    help_category = "System"
    directory = settings.GAME_DIR
    repo_type = "game"
    remote_link = "[your remote link]"


# CmdSet for easily install all commands
class GitCmdSet(CmdSet):
    """
    The git command.
    """

    def at_cmdset_creation(self):
        self.add(CmdGit)
        self.add(CmdGitEvennia)



#
# This sets up how models are displayed
# in the web admin interface.
#
from django import forms
from django.conf import settings
from django.contrib import admin, messages
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.admin.utils import unquote
from django.contrib.admin.widgets import FilteredSelectMultiple, ForeignKeyRawIdWidget
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.forms import UserChangeForm, UserCreationForm
from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import path, reverse
from django.utils.decorators import method_decorator
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.views.decorators.debug import sensitive_post_parameters

from evennia.accounts.models import AccountDB
from evennia.objects.models import ObjectDB
from evennia.utils import create

from . import utils as adminutils
from .attributes import AttributeInline
from .tags import TagInline

sensitive_post_parameters_m = method_decorator(sensitive_post_parameters())


# handle the custom User editor
class AccountChangeForm(UserChangeForm):
    """
    Modify the accountdb class.

    """

    class Meta(object):
        model = AccountDB
        fields = "__all__"

    username = forms.RegexField(
        label="Username",
        max_length=30,
        regex=r"^[\w. @+-]+$",
        widget=forms.TextInput(attrs={"size": "30"}),
        error_messages={
            "invalid": "This value may contain only letters, spaces, numbers "
            "and @/./+/-/_ characters."
        },
        help_text="30 characters or fewer. Letters, spaces, digits and " "@/./+/-/_ only.",
    )

    db_typeclass_path = forms.ChoiceField(
        label="Typeclass",
        help_text="This is the Python-path to the class implementing the actual account functionality. "
        "You usually don't need to change this from the default.<BR>"
        "If your custom class is not found here, it may not be imported into Evennia yet.",
        choices=lambda: adminutils.get_and_load_typeclasses(parent=AccountDB),
    )

    db_lock_storage = forms.CharField(
        label="Locks",
        required=False,
        widget=forms.Textarea(attrs={"cols": "100", "rows": "2"}),
        help_text="Locks limit access to the entity. Written on form `type:lockdef;type:lockdef..."
        "<BR>(Permissions (used with the perm() lockfunc) are Tags with the 'permission' type)",
    )

    db_cmdset_storage = forms.CharField(
        label="CommandSet",
        initial=settings.CMDSET_ACCOUNT,
        widget=forms.TextInput(attrs={"size": "78"}),
        required=False,
    )

    is_superuser = forms.BooleanField(
        label="Superuser status",
        required=False,
        help_text="Superusers bypass all in-game locks and has all "
        "permissions without explicitly assigning them. Usually "
        "only one superuser (user #1) is needed and only a superuser "
        "can create another superuser.<BR>"
        "Only Superusers can change the user/group permissions below.",
    )

    def clean_username(self):
        """
        Clean the username and check its existence.

        """
        username = self.cleaned_data["username"]
        if username.upper() == self.instance.username.upper():
            return username
        elif AccountDB.objects.filter(username__iexact=username):
            raise forms.ValidationError("An account with that name " "already exists.")
        return self.cleaned_data["username"]

    def __init__(self, *args, **kwargs):
        """
        Tweak some fields dynamically.

        """
        super().__init__(*args, **kwargs)

        # better help text for cmdset_storage
        account_cmdset = settings.CMDSET_ACCOUNT
        self.fields["db_cmdset_storage"].help_text = (
            "Path to Command-set path. Most non-character objects don't need a cmdset"
            " and can leave this field blank. Default cmdset-path<BR> for Accounts "
            f"is <strong>{account_cmdset}</strong> ."
        )


class AccountCreationForm(UserCreationForm):
    """
    Create a new AccountDB instance.
    """

    class Meta(object):
        model = AccountDB
        fields = "__all__"

    username = forms.RegexField(
        label="Username",
        max_length=30,
        regex=r"^[\w. @+-]+$",
        widget=forms.TextInput(attrs={"size": "30"}),
        error_messages={
            "invalid": "This value may contain only letters, spaces, numbers "
            "and @/./+/-/_ characters."
        },
        help_text="30 characters or fewer. Letters, spaces, digits and " "@/./+/-/_ only.",
    )

    def clean_username(self):
        """
        Cleanup username.
        """
        username = self.cleaned_data["username"]
        if AccountDB.objects.filter(username__iexact=username):
            raise forms.ValidationError("An account with that name already " "exists.")
        return username


class AccountTagInline(TagInline):
    """
    Inline Account Tags.

    """

    model = AccountDB.db_tags.through
    related_field = "accountdb"


class AccountAttributeInline(AttributeInline):
    """
    Inline Account Attributes.

    """

    model = AccountDB.db_attributes.through
    related_field = "accountdb"


class ObjectPuppetInline(admin.StackedInline):
    """
    Inline creation of puppet-Object in Account.

    """

    from .objects import ObjectCreateForm

    verbose_name = "Puppeted Object"
    model = ObjectDB
    view_on_site = False
    show_change_link = True
    # template = "admin/accounts/stacked.html"
    form = ObjectCreateForm
    fieldsets = (
        (
            None,
            {
                "fields": (
                    ("db_key", "db_typeclass_path"),
                    ("db_location", "db_home", "db_destination"),
                    "db_cmdset_storage",
                    "db_lock_storage",
                ),
                "description": "Object currently puppeted by the account (note that this "
                "will go away if account logs out or unpuppets)",
            },
        ),
    )

    extra = 0
    readonly_fields = (
        "db_key",
        "db_typeclass_path",
        "db_destination",
        "db_location",
        "db_home",
        "db_account",
        "db_cmdset_storage",
        "db_lock_storage",
    )

    # disable adding/deleting this inline - read-only!
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(AccountDB)
class AccountAdmin(BaseUserAdmin):
    """
    This is the main creation screen for Users/accounts

    """

    list_display = (
        "id",
        "username",
        "is_staff",
        "is_superuser",
        "db_typeclass_path",
        "db_date_created",
    )
    list_display_links = ("id", "username")
    form = AccountChangeForm
    add_form = AccountCreationForm
    search_fields = ["=id", "^username", "db_typeclass_path"]
    ordering = ["-db_date_created", "id"]
    list_filter = ["is_superuser", "is_staff", "db_typeclass_path"]
    inlines = [AccountTagInline, AccountAttributeInline]
    readonly_fields = ["db_date_created", "serialized_string", "puppeted_objects"]
    view_on_site = False
    fieldsets = (
        (
            None,
            {
                "fields": (
                    ("username", "db_typeclass_path"),
                    "password",
                    "email",
                    "db_date_created",
                    "db_lock_storage",
                    "db_cmdset_storage",
                    "puppeted_objects",
                    "serialized_string",
                )
            },
        ),
        (
            "Admin/Website properties",
            {
                "fields": (
                    ("first_name", "last_name"),
                    "last_login",
                    "date_joined",
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "user_permissions",
                    "groups",
                ),
                "description": "<i>Used by the website/Django admin. "
                "Except for `superuser status`, the permissions are not used in-game.</i>",
            },
        ),
    )

    add_fieldsets = (
        (
            None,
            {
                "fields": ("username", "password1", "password2", "email"),
                "description": "<i>These account details are shared by the admin "
                "system and the game.</i>",
            },
        ),
    )

    def serialized_string(self, obj):
        """
        Get the serialized version of the object.

        """
        from evennia.utils import dbserialize

        return str(dbserialize.pack_dbobj(obj))

    serialized_string.help_text = (
        "Copy & paste this string into an Attribute's `value` field to store this account there."
    )

    def puppeted_objects(self, obj):
        """
        Get any currently puppeted objects (read only list)

        """
        return mark_safe(
            ", ".join(
                '<a href="{url}">{name}</a>'.format(
                    url=reverse("admin:objects_objectdb_change", args=[obj.id]), name=obj.db_key
                )
                for obj in ObjectDB.objects.filter(db_account=obj)
            )
        )

    puppeted_objects.help_text = (
        "Objects currently puppeted by this Account. "
        "Link new ones from the `Objects` admin page.<BR>"
        "Note that these will disappear when a user unpuppets or goes offline - "
        "this is normal."
    )

    def get_form(self, request, obj=None, **kwargs):
        """
        Overrides help texts.

        """
        help_texts = kwargs.get("help_texts", {})
        help_texts["serialized_string"] = self.serialized_string.help_text
        help_texts["puppeted_objects"] = self.puppeted_objects.help_text
        kwargs["help_texts"] = help_texts

        # security disabling for non-superusers
        form = super().get_form(request, obj, **kwargs)
        disabled_fields = set()
        if not request.user.is_superuser:
            disabled_fields |= {"is_superuser", "user_permissions", "user_groups"}
        for field_name in disabled_fields:
            if field_name in form.base_fields:
                form.base_fields[field_name].disabled = True
        return form

    @sensitive_post_parameters_m
    def user_change_password(self, request, id, form_url=""):
        user = self.get_object(request, unquote(id))
        if not self.has_change_permission(request, user):
            raise PermissionDenied
        if user is None:
            raise Http404("%(name)s object with primary key %(key)r does not exist.") % {
                "name": self.model._meta.verbose_name,
                "key": escape(id),
            }
        if request.method == "POST":
            form = self.change_password_form(user, request.POST)
            if form.is_valid():
                form.save()
                change_message = self.construct_change_message(request, form, None)
                self.log_change(request, user, change_message)
                msg = "Password changed successfully."
                messages.success(request, msg)
                update_session_auth_hash(request, form.user)
                return HttpResponseRedirect(
                    reverse(
                        "%s:%s_%s_change"
                        % (
                            self.admin_site.name,
                            user._meta.app_label,
                            # the model_name is something we need to hardcode
                            # since our accountdb is a proxy:
                            "accountdb",
                        ),
                        args=(user.pk,),
                    )
                )
        else:
            form = self.change_password_form(user)

        fieldsets = [(None, {"fields": list(form.base_fields)})]
        adminForm = admin.helpers.AdminForm(form, fieldsets, {})

        context = {
            "title": "Change password: %s" % escape(user.get_username()),
            "adminForm": adminForm,
            "form_url": form_url,
            "form": form,
            "is_popup": (IS_POPUP_VAR in request.POST or IS_POPUP_VAR in request.GET),
            "add": True,
            "change": False,
            "has_delete_permission": False,
            "has_change_permission": True,
            "has_absolute_url": False,
            "opts": self.model._meta,
            "original": user,
            "save_as": False,
            "show_save": True,
            **self.admin_site.each_context(request),
        }

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request,
            self.change_user_password_template or "admin/auth/user/change_password.html",
            context,
        )

    def save_model(self, request, obj, form, change):
        """
        Custom save actions.

        Args:
            request (Request): Incoming request.
            obj (Object): Object to save.
            form (Form): Related form instance.
            change (bool): False if this is a new save and not an update.

        """
        obj.save()
        if not change:
            # calling hooks for new account
            obj.set_class_from_typeclass(typeclass_path=settings.BASE_ACCOUNT_TYPECLASS)
            obj.basetype_setup()
            obj.at_account_creation()

    def response_add(self, request, obj, post_url_continue=None):
        from django.http import HttpResponseRedirect
        from django.urls import reverse

        return HttpResponseRedirect(reverse("admin:accounts_accountdb_change", args=[obj.id]))
